from uuid import UUID

import pytest

from supriya.patterns import (
    CompositeEvent,
    Event,
    EventPattern,
    GroupAllocateEvent,
    GroupPattern,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    SequencePattern,
)

from .conftest import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, inner_pattern, release_time, expected, is_infinite",
    [
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            0.0,
            [
                CompositeEvent([GroupAllocateEvent(UUID(int=0), delta=0.0)]),
                NoteEvent(UUID(int=1), a=1, target_node=UUID(int=0)),
                NoteEvent(UUID(int=2), a=2, target_node=UUID(int=0)),
                CompositeEvent([NodeFreeEvent(UUID(int=0), delta=0.0)]),
            ],
            False,
        ),
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            0.25,
            [
                CompositeEvent([GroupAllocateEvent(UUID(int=0), delta=0.0)]),
                NoteEvent(UUID(int=1), a=1, target_node=UUID(int=0)),
                NoteEvent(UUID(int=2), a=2, target_node=UUID(int=0)),
                CompositeEvent(
                    [NullEvent(delta=0.25), NodeFreeEvent(UUID(int=0), delta=0.0)]
                ),
            ],
            False,
        ),
    ],
)
def test_pattern(
    stop_at: float | None,
    inner_pattern: EventPattern,
    release_time: float,
    expected: list[Event],
    is_infinite: bool,
) -> None:
    pattern = GroupPattern(inner_pattern, release_time=release_time)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



from supriya.contexts.allocators import BlockAllocator


def test_allocate() -> None:
    allocator = BlockAllocator(heap_minimum=0, heap_maximum=16)
    assert allocator.allocate(4) == 0
    assert allocator.allocate(4) == 4
    assert allocator.allocate(4) == 8
    assert allocator.allocate(4) == 12
    assert allocator.allocate(4) is None
    allocator.free(0)
    allocator.free(4)
    allocator.free(8)
    allocator.free(12)
    assert allocator.allocate(20) is None
    assert allocator.allocate(4) == 0
    assert allocator.allocate(4) == 4
    allocator.free(4)
    assert allocator.allocate(1) == 4
    assert allocator.allocate(1) == 5



import bisect
import dataclasses
import threading


@dataclasses.dataclass(order=True)
class Block:
    start_offset: int = 0
    stop_offset: int = 0

    @property
    def size(self) -> float:
        return self.stop_offset - self.start_offset


class BlockAllocator:
    """
    A block allocator.

    ::

        >>> from supriya.contexts.allocators import BlockAllocator
        >>> allocator = BlockAllocator(heap_maximum=16)

    ::

        >>> allocator.allocate(4)
        0

    ::

        >>> allocator.allocate(4)
        4

    ::

        >>> allocator.allocate(4)
        8

    ::

        >>> allocator.allocate(8) is None
        True

    ::

        >>> allocator.free(8)
        >>> allocator.allocate(8)
        8
    """

    ### INITIALIZER ###

    def __init__(self, heap_maximum: int = 1024, heap_minimum: int = 0) -> None:
        self._used_dict: dict[int, Block] = {}
        self._free_heap: list[Block] = [
            Block(
                start_offset=heap_minimum,
                stop_offset=heap_maximum,
            )
        ]
        self._lock = threading.Lock()

    ### SPECIAL METHODS ###

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state["_lock"]
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self._lock = threading.Lock()

    ### PUBLIC METHODS ###

    def allocate(self, size: int = 1) -> int | None:
        with self._lock:
            for i, block in enumerate(self._free_heap):
                if size <= block.size:
                    break
            else:
                return None
            if size < block.size:
                self._free_heap[i] = Block(
                    start_offset=block.start_offset + size,
                    stop_offset=block.stop_offset,
                )
            else:
                self._free_heap.pop(i)
            self._used_dict[block.start_offset] = Block(
                start_offset=block.start_offset,
                stop_offset=block.start_offset + size,
            )
            return int(block.start_offset)

    def free(self, index: int) -> None:
        with self._lock:
            if (block := self._used_dict.pop(index, None)) is None:
                return
            self._free_heap.insert(
                index := bisect.bisect(self._free_heap, block), block
            )
            # coalesce right
            if (
                index < (len(self._free_heap) - 1)
                and self._free_heap[index].stop_offset
                == self._free_heap[index + 1].start_offset
            ):
                self._free_heap[index].stop_offset = self._free_heap.pop(
                    index + 1
                ).stop_offset
            # coalesce left
            if (
                index > 0
                and self._free_heap[index].start_offset
                == self._free_heap[index - 1].stop_offset
            ):
                self._free_heap[index - 1].stop_offset = self._free_heap.pop(
                    index
                ).stop_offset


class NodeIdAllocator:
    """
    A node ID allocator.

    ::

        >>> from supriya.contexts.allocators import NodeIdAllocator
        >>> allocator = NodeIdAllocator()
        >>> for _ in range(3):
        ...     allocator.allocate_node_id()
        ...
        1000
        1001
        1002

    ::

        >>> for _ in range(3):
        ...     allocator.allocate_permanent_node_id()
        ...
        1
        2
        3

    ::

        >>> allocator.free_permanent_node_id(2)

    ::

        >>> allocator.allocate_permanent_node_id()
        2
    """

    ### INITIALIZER ###

    def __init__(
        self, client_id: int = 0, initial_node_id: int = 1000, locked=True
    ) -> None:
        if client_id > 31:
            raise ValueError
        self._initial_node_id = initial_node_id
        self._client_id = client_id
        self._mask = self._client_id << 26
        self._temp = self._initial_node_id
        self._next_permanent_id = 1
        self._freed_permanent_ids: set[int] = set()
        self._lock = threading.Lock()

    ### SPECIAL METHODS ###

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state["_lock"]
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self._lock = threading.Lock()

    ### PUBLIC METHODS ###

    def allocate(self, count: int = 1) -> int:
        return self.allocate_node_id(count)

    def allocate_node_id(self, count: int = 1) -> int:
        with self._lock:
            x = self._temp
            temp = x + count
            if 0x03FFFFFF < temp:
                temp = (temp % 0x03FFFFFF) + self._initial_node_id
            self._temp = temp
            x = x | self._mask
            return x

    def allocate_permanent_node_id(self) -> int:
        with self._lock:
            if self._freed_permanent_ids:
                x = min(self._freed_permanent_ids)
                self._freed_permanent_ids.remove(x)
            else:
                x = self._next_permanent_id
                self._next_permanent_id = min(x + 1, self._initial_node_id - 1)
            x = x | self._mask
            return x

    def free(self, node_id: int) -> None:
        if node_id < self._initial_node_id:
            self.free_permanent_node_id(node_id)

    def free_permanent_node_id(self, node_id: int) -> None:
        with self._lock:
            node_id = node_id & 0x03FFFFFF
            if node_id < self._initial_node_id:
                self._freed_permanent_ids.add(node_id)



from libcpp cimport bool


cdef extern from "scope_buffer.hpp" namespace "detail_server_shm":
    cdef cppclass scope_buffer_reader:
        bool valid()
        bool pull(unsigned int& frames)
        float* data()
        unsigned int max_frames()
        unsigned int channels()


cdef extern from "server_shm.hpp" namespace "detail_server_shm":
    cdef cppclass server_shared_memory_client:
        server_shared_memory_client(unsigned int port_number) except +
        float* get_control_busses()
        void set_control_bus(int bus, float value)
        scope_buffer_reader get_scope_buffer_reader(unsigned int index)



import asyncio
import collections
import concurrent.futures
import contextlib
import dataclasses
import datetime
import logging
import pprint
import socket
import socketserver
import struct
import threading
import time
from collections.abc import Sequence as SequenceABC
from enum import Enum
from queue import Empty, Queue
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    Iterator,
    Literal,
    NamedTuple,
    Sequence,
    TypeAlias,
    Union,
    cast,
)

from .enums import BootStatus
from .typing import FutureLike, SupportsOsc
from .ugens import decompile_synthdefs
from .utils import group_by_count

osc_protocol_logger = logging.getLogger(__name__)
osc_in_logger = logging.getLogger("supriya.osc.in")
osc_out_logger = logging.getLogger("supriya.osc.out")
udp_in_logger = logging.getLogger("supriya.udp.in")
udp_out_logger = logging.getLogger("supriya.udp.out")


BUNDLE_PREFIX = b"#bundle\x00"
IMMEDIATELY = struct.pack(">Q", 1)
NTP_TIMESTAMP_TO_SECONDS = 1.0 / 2.0**32.0
SECONDS_TO_NTP_TIMESTAMP = 2.0**32.0
SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
NTP_EPOCH = datetime.date(1900, 1, 1)
NTP_DELTA = (SYSTEM_EPOCH - NTP_EPOCH).days * 24 * 3600


OscAddress: TypeAlias = Enum | int | str
OscArgument: TypeAlias = Union[
    Enum,
    "OscBundle",
    "OscMessage",
    SequenceABC["OscArgument"],
    bool,
    bytes,
    float,
    str,
    None,
]


def format_datagram(datagram: bytes) -> str:
    result: list[str] = ["size {}".format(len(datagram))]
    index = 0
    while index < len(datagram):
        hex_blocks = []
        ascii_block = ""
        for chunk in group_by_count(datagram[index : index + 16], 4):
            hex_block = []
            for byte in chunk:
                if 31 < int(byte) < 127:
                    char = chr(int(byte))
                else:
                    char = "."
                ascii_block += char
                hexed = hex(byte)[2:].zfill(2)
                hex_block.append(hexed)
            hex_blocks.append(" ".join(hex_block))
        line = "{: >4}   ".format(index)
        line += "{: <53}".format("  ".join(hex_blocks))
        line += "|{}|".format(ascii_block)
        result.append(line)
        index += 16
    return "\n".join(result)


class OscMessage:
    """
    An OSC message.

    .. container:: example

        ::

            >>> from supriya.osc import OscMessage
            >>> osc_message = OscMessage("/g_new", 0, 0)
            >>> osc_message
            OscMessage('/g_new', 0, 0)

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/g_new', 0, 0)

        ::

            >>> print(osc_message)
            size 20
               0   2f 67 5f 6e  65 77 00 00  2c 69 69 00  00 00 00 00   |/g_new..,ii.....|
              16   00 00 00 00                                          |....|

    .. container:: example

        ::

            >>> osc_message = OscMessage("/foo", True, [None, [3.25]], OscMessage("/bar"))
            >>> osc_message
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> print(osc_message)
            size 40
               0   2f 66 6f 6f  00 00 00 00  2c 54 5b 4e  5b 66 5d 5d   |/foo....,T[N[f]]|
              16   62 00 00 00  40 50 00 00  00 00 00 0c  2f 62 61 72   |b...@P....../bar|
              32   00 00 00 00  2c 00 00 00                             |....,...|

    .. container:: example

        ::

            >>> osc_message = supriya.osc.OscMessage(
            ...     "/foo",
            ...     1,
            ...     2.5,
            ...     supriya.osc.OscBundle(
            ...         contents=(
            ...             supriya.osc.OscMessage("/bar", "baz", 3.0),
            ...             supriya.osc.OscMessage("/ffff", False, True, None),
            ...         )
            ...     ),
            ...     ["a", "b", ["c", "d"]],
            ... )
            >>> osc_message
            OscMessage('/foo', 1, 2.5, OscBundle(contents=[OscMessage('/bar', 'baz', 3.0), OscMessage('/ffff', False, True, None)]), ['a', 'b', ['c', 'd']])

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', 1, 2.5, OscBundle(contents=[OscMessage('/bar', 'baz', 3.0), OscMessage('/ffff', False, True, None)]), ['a', 'b', ['c', 'd']])

        ::

            >>> print(osc_message)
            size 112
               0   2f 66 6f 6f  00 00 00 00  2c 69 66 62  5b 73 73 5b   |/foo....,ifb[ss[|
              16   73 73 5d 5d  00 00 00 00  00 00 00 01  40 20 00 00   |ss]]........@ ..|
              32   00 00 00 3c  23 62 75 6e  64 6c 65 00  00 00 00 00   |...<#bundle.....|
              48   00 00 00 01  00 00 00 14  2f 62 61 72  00 00 00 00   |......../bar....|
              64   2c 73 66 00  62 61 7a 00  40 40 00 00  00 00 00 10   |,sf.baz.@@......|
              80   2f 66 66 66  66 00 00 00  2c 46 54 4e  00 00 00 00   |/ffff...,FTN....|
              96   61 00 00 00  62 00 00 00  63 00 00 00  64 00 00 00   |a...b...c...d...|
    """

    ### INITIALIZER ###

    def __init__(self, address: OscAddress, *contents: OscArgument) -> None:
        if isinstance(address, Enum):
            address = address.value
        if not isinstance(address, (str, int)):
            raise ValueError(f"address must be int or str, got {address}")
        self.address = address
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.address != other.address:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(repr(_) for _ in [self.address, *self.contents]),
        )

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_blob(data: bytes) -> tuple[bytes, bytes]:
        actual_length, remainder = struct.unpack(">I", data[:4])[0], data[4:]
        padded_length = actual_length
        if actual_length % 4 != 0:
            padded_length = (actual_length // 4 + 1) * 4
        return remainder[:padded_length][:actual_length], remainder[padded_length:]

    @staticmethod
    def _decode_string(data: bytes) -> tuple[str, bytes]:
        actual_length = data.index(b"\x00")
        padded_length = (actual_length // 4 + 1) * 4
        return str(data[:actual_length], "ascii"), data[padded_length:]

    @staticmethod
    def _encode_string(value: str) -> bytes:
        result = bytes(value + "\x00", "ascii")
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @staticmethod
    def _encode_blob(value: bytes) -> bytes:
        result = bytes(struct.pack(">I", len(value)) + value)
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @classmethod
    def _encode_value(cls, value: OscArgument) -> tuple[str, bytes]:
        type_tags, encoded_value = "", b""
        if isinstance(value, Enum):
            if isinstance(value.value, str):
                type_tags += "s"
                encoded_value = cls._encode_string(value.value)
            else:
                type_tags += "i"
                encoded_value += struct.pack(">i", value.value)
        elif isinstance(value, (OscBundle, OscMessage)):
            type_tags += "b"
            encoded_value = cls._encode_blob(value.to_datagram())
        elif isinstance(value, (bytearray, bytes)):
            type_tags += "b"
            encoded_value = cls._encode_blob(value)
        elif isinstance(value, str):
            type_tags += "s"
            encoded_value = cls._encode_string(value)
        elif isinstance(value, bool):
            type_tags += "T" if value else "F"
        elif isinstance(value, float):
            type_tags += "f"
            encoded_value += struct.pack(">f", value)
        elif isinstance(value, int):
            type_tags += "i"
            encoded_value += struct.pack(">i", value)
        elif value is None:
            type_tags += "N"
        elif isinstance(value, SequenceABC):
            type_tags += "["
            for sub_value in value:
                sub_type_tags, sub_encoded_value = cls._encode_value(sub_value)
                type_tags += sub_type_tags
                encoded_value += sub_encoded_value
            type_tags += "]"
        else:
            message = "Cannot encode {!r}".format(value)
            raise TypeError(message)
        return type_tags, encoded_value

    ### PUBLIC METHODS ###

    def to_datagram(self) -> bytes:
        # address can be a string or (in SuperCollider) an int
        if isinstance(self.address, str):
            encoded_address = self._encode_string(self.address)
        else:
            encoded_address = struct.pack(">i", self.address)
        encoded_type_tags = ","
        encoded_contents = b""
        for value in self.contents or ():
            type_tags, encoded_value = self._encode_value(value)
            encoded_type_tags += type_tags
            encoded_contents += encoded_value
        return (
            encoded_address + self._encode_string(encoded_type_tags) + encoded_contents
        )

    @classmethod
    def from_datagram(cls, datagram: bytes) -> "OscMessage":
        remainder = datagram
        address, remainder = cls._decode_string(remainder)
        type_tags, remainder = cls._decode_string(remainder)
        contents: list[OscArgument] = []
        array_stack = [contents]
        for type_tag in type_tags[1:]:
            if type_tag == "i":
                value, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "f":
                value, remainder = struct.unpack(">f", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "d":
                value, remainder = struct.unpack(">d", remainder[:8])[0], remainder[8:]
                array_stack[-1].append(value)
            elif type_tag == "s":
                value, remainder = cls._decode_string(remainder)
                array_stack[-1].append(value)
            elif type_tag == "b":
                value, remainder = cls._decode_blob(remainder)
                for class_ in (OscBundle, OscMessage):
                    try:
                        value = class_.from_datagram(value)
                        break
                    except Exception:
                        pass
                array_stack[-1].append(value)
            elif type_tag == "T":
                array_stack[-1].append(True)
            elif type_tag == "F":
                array_stack[-1].append(False)
            elif type_tag == "N":
                array_stack[-1].append(None)
            elif type_tag == "[":
                array: list[OscArgument] = []
                array_stack[-1].append(array)
                array_stack.append(array)
            elif type_tag == "]":
                array_stack.pop()
            else:
                raise RuntimeError(f"Unable to parse type {type_tag!r}")
        return cls(address, *contents)

    def to_list(self):
        result = [self.address]
        for x in self.contents:
            if hasattr(x, "to_list"):
                result.append(x.to_list())
            else:
                result.append(x)
        return result

    def to_osc(self) -> "OscMessage":
        return self


class OscBundle:
    """
    An OSC bundle.

    ::

        >>> import supriya.osc
        >>> message_one = supriya.osc.OscMessage("/one", 1)
        >>> message_two = supriya.osc.OscMessage("/two", 2)
        >>> message_three = supriya.osc.OscMessage("/three", 3)

    ::

        >>> inner_bundle = supriya.osc.OscBundle(
        ...     timestamp=1401557034.5,
        ...     contents=(message_one, message_two),
        ... )
        >>> inner_bundle
        OscBundle(timestamp=1401557034.5, contents=[OscMessage('/one', 1), OscMessage('/two', 2)])

    ::

        >>> print(inner_bundle)
        size 56
           0   23 62 75 6e  64 6c 65 00  d7 34 8e aa  80 00 00 00   |#bundle..4......|
          16   00 00 00 10  2f 6f 6e 65  00 00 00 00  2c 69 00 00   |..../one....,i..|
          32   00 00 00 01  00 00 00 10  2f 74 77 6f  00 00 00 00   |......../two....|
          48   2c 69 00 00  00 00 00 02                             |,i......|

    ::

        >>> outer_bundle = supriya.osc.OscBundle(
        ...     contents=(inner_bundle, message_three),
        ... )
        >>> outer_bundle
        OscBundle(contents=[OscBundle(timestamp=1401557034.5, contents=[OscMessage('/one', 1), OscMessage('/two', 2)]), OscMessage('/three', 3)])

    ::

        >>> print(outer_bundle)
        size 96
           0   23 62 75 6e  64 6c 65 00  00 00 00 00  00 00 00 01   |#bundle.........|
          16   00 00 00 38  23 62 75 6e  64 6c 65 00  d7 34 8e aa   |...8#bundle..4..|
          32   80 00 00 00  00 00 00 10  2f 6f 6e 65  00 00 00 00   |......../one....|
          48   2c 69 00 00  00 00 00 01  00 00 00 10  2f 74 77 6f   |,i........../two|
          64   00 00 00 00  2c 69 00 00  00 00 00 02  00 00 00 10   |....,i..........|
          80   2f 74 68 72  65 65 00 00  2c 69 00 00  00 00 00 03   |/three..,i......|

    ::

        >>> datagram = outer_bundle.to_datagram()

    ::

        >>> decoded_bundle = supriya.osc.OscBundle.from_datagram(datagram)
        >>> decoded_bundle
        OscBundle(contents=[OscBundle(timestamp=1401557034.5, contents=[OscMessage('/one', 1), OscMessage('/two', 2)]), OscMessage('/three', 3)])

    ::

        >>> decoded_bundle == outer_bundle
        True
    """

    ### INITIALIZER ###

    def __init__(
        self,
        timestamp: float | None = None,
        *,
        contents: SequenceABC[Union["OscBundle", "OscMessage"]],
    ) -> None:
        prototype = (OscMessage, type(self))
        self.timestamp = timestamp
        contents = contents or ()
        for x in contents or ():
            if not isinstance(x, prototype):
                raise ValueError(contents)
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.timestamp != other.timestamp:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        parts = ["{}(".format(type(self).__name__)]
        if self.timestamp is not None:
            parts.append(f"timestamp={self.timestamp}")
            if self.contents:
                parts.append(", ")
        if self.contents:
            parts.append(f"contents={list(self.contents)!r}")
        parts.append(")")
        return "".join(parts)

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_date(data: bytes) -> tuple[float | None, bytes]:
        data, remainder = data[:8], data[8:]
        if data == IMMEDIATELY:
            return None, remainder
        date = (struct.unpack(">Q", data)[0] / SECONDS_TO_NTP_TIMESTAMP) - NTP_DELTA
        return date, remainder

    @staticmethod
    def _encode_date(seconds: float | None, realtime: bool = True) -> bytes:
        if seconds is None:
            return IMMEDIATELY
        if realtime:
            seconds = seconds + NTP_DELTA
        if seconds >= 4294967296:  # 2**32
            seconds = seconds % 4294967296
        return struct.pack(">Q", int(seconds * SECONDS_TO_NTP_TIMESTAMP))

    ### PUBLIC METHODS ###

    @classmethod
    def from_datagram(cls, datagram: bytes) -> "OscBundle":
        if not datagram.startswith(BUNDLE_PREFIX):
            raise ValueError("datagram is not a bundle")
        remainder = datagram[8:]
        timestamp, remainder = cls._decode_date(remainder)
        contents = []
        while len(remainder):
            length, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
            item: OscBundle | OscMessage
            if remainder.startswith(BUNDLE_PREFIX):
                item = cls.from_datagram(remainder[:length])
            else:
                item = OscMessage.from_datagram(remainder[:length])
            contents.append(item)
            remainder = remainder[length:]
        osc_bundle = cls(timestamp=timestamp, contents=tuple(contents))
        return osc_bundle

    @classmethod
    def partition(
        cls, messages: list[OscMessage], timestamp: float | None = None
    ) -> list["OscBundle"]:
        bundles = []
        contents = []
        messages_ = collections.deque(messages)
        remaining = maximum = 8192 - len(BUNDLE_PREFIX) - 4
        while messages_:
            message = messages_.popleft()
            datagram = message.to_datagram()
            remaining -= len(datagram) + 4
            if remaining > 0:
                contents.append(message)
            else:
                bundles.append(cls(timestamp=timestamp, contents=contents))
                contents = [message]
                remaining = maximum
        if contents:
            bundles.append(cls(timestamp=timestamp, contents=contents))
        return bundles

    def to_datagram(self, realtime: bool = True) -> bytes:
        datagram: bytes = BUNDLE_PREFIX
        datagram += self._encode_date(self.timestamp, realtime=realtime)
        for content in self.contents:
            content_datagram = content.to_datagram()
            datagram += struct.pack(">i", len(content_datagram))
            datagram += content_datagram
        return datagram

    def to_list(self):
        result = [self.timestamp]
        result.append([x.to_list() for x in self.contents])
        return result

    def to_osc(self) -> "OscBundle":
        return self


def format_messages(messages: Sequence[OscBundle | OscMessage]) -> str:
    """
    Format a sequence of OSC messages as a string.

    Provides a more concise means of verifying OSC contents in the test suite
    than comparing the messages directly.
    """

    def sanitize(list_):
        for i, x in enumerate(list_):
            if isinstance(x, bytes):
                try:
                    decompiled = decompile_synthdefs(x)
                    list_[i] = decompiled[0] if len(decompiled) == 1 else decompiled
                except Exception:
                    pass
            elif isinstance(x, list):
                sanitize(x)
        return list_

    lines: list[str] = []
    for message in messages:
        sanitized = sanitize(message.to_list())
        formatted = pprint.pformat(sanitized, width=120)
        for i, line in enumerate(formatted.splitlines()):
            if i == 0:
                prefix = "-"
            else:
                prefix = " "
            lines.append(f"{prefix} {line}")
    return "\n".join(lines)


def find_free_port() -> int:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class OscProtocolOffline(Exception):
    pass


class OscProtocolAlreadyConnected(Exception):
    pass


class OscCallback(NamedTuple):
    protocol: "OscProtocol"
    pattern: tuple[str | int | float, ...]
    procedure: Callable
    failure_pattern: tuple[float | int | str, ...] | None = None
    once: bool = False
    args: tuple | None = None
    kwargs: dict | None = None

    def unregister(self) -> None:
        self.protocol.unregister(self)


@dataclasses.dataclass
class HealthCheck:
    request_pattern: list[str]
    response_pattern: list[str]
    active: bool = True
    timeout: float = 1.0
    backoff_factor: float = 1.5
    max_attempts: int = 5


class CaptureEntry(NamedTuple):
    timestamp: float
    label: Literal["R", "S"]
    message: OscBundle | OscMessage
    raw_message: SequenceABC | SupportsOsc | str | None = None


class Capture:
    ### INITIALIZER ###

    def __init__(self, osc_protocol: "OscProtocol") -> None:
        self.osc_protocol = osc_protocol
        self.entries: list[CaptureEntry] = []

    ### SPECIAL METHODS ###

    def __enter__(self) -> "Capture":
        self.osc_protocol.captures.add(self)
        self.entries[:] = []
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.osc_protocol.captures.remove(self)

    def __getitem__(self, i: int | slice) -> CaptureEntry | list[CaptureEntry]:
        return self.entries[i]

    def __iter__(self) -> Iterator[CaptureEntry]:
        return iter(self.entries)

    def __len__(self) -> int:
        return len(self.entries)

    ### PUBLIC METHODS ###

    def add_entry(
        self,
        timestamp: float,
        label: Literal["R", "S"],
        message: OscBundle | OscMessage,
        raw_message: SequenceABC | SupportsOsc | str | None = None,
    ) -> None:
        self.entries.append(
            CaptureEntry(
                timestamp=timestamp,
                label=label,
                message=message,
                raw_message=raw_message,
            )
        )

    def filtered(
        self, sent: bool = True, received: bool = True, status: bool = False
    ) -> list[CaptureEntry]:
        entries = []
        for entry in self.entries:
            if entry.label == "R" and not received:
                continue
            if entry.label == "S" and not sent:
                continue
            if (
                not status
                and isinstance(entry.message, OscMessage)
                and entry.message.address in ("/status", "/status.reply")
            ):
                continue
            entries.append(entry)
        return entries


class OscProtocol:
    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: str | None = None,
        on_connect_callback: Callable | None = None,
        on_disconnect_callback: Callable | None = None,
        on_panic_callback: Callable | None = None,
    ) -> None:
        self.callbacks: dict[Any, Any] = {}
        self.captures: set[Capture] = set()
        self.healthcheck: HealthCheck | None = None
        self.healthcheck_osc_callback: OscCallback | None = None
        self.attempts = 0
        self.ip_address = "127.0.0.1"
        self.name = name
        self.port = 57551
        self.on_connect_callback = on_connect_callback
        self.on_disconnect_callback = on_disconnect_callback
        self.on_panic_callback = on_panic_callback
        self.status = BootStatus.OFFLINE

    ### PRIVATE METHODS ###

    def _activate_healthcheck(self) -> bool:
        if not self.healthcheck:
            return False
        elif self.healthcheck.active:
            return False
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "activating healthcheck..."
        )
        return True

    def _add_callback(self, callback: OscCallback) -> None:
        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            callback_map = self.callbacks
            for item in pattern:
                callbacks, callback_map = callback_map.setdefault(item, ([], {}))
            callbacks.append(callback)

    def _disconnect(self, panicked: bool = False) -> Awaitable[None] | None:
        if panicked:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "panicking ..."
            )
        else:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "disconnecting ..."
            )
        self.status = BootStatus.QUITTING
        if self.healthcheck_osc_callback is not None:
            self.unregister(self.healthcheck_osc_callback)
        return None

    def _match_callbacks(self, message) -> list[OscCallback]:
        items = (message.address,) + message.contents
        matching_callbacks = []
        callback_map = self.callbacks
        for item in items:
            if item not in callback_map:
                break
            callbacks, callback_map = callback_map[item]
            matching_callbacks.extend(callbacks)
        for callback in matching_callbacks:
            if callback.once:
                self.unregister(callback)
        return matching_callbacks

    def _on_connect(self, *, boot_future: FutureLike[bool]) -> Awaitable[None] | None:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "... connected!"
        )
        self.status = BootStatus.ONLINE
        boot_future.set_result(True)
        return None

    def _on_disconnect(
        self,
        *,
        boot_future: FutureLike[bool],
        exit_future: FutureLike[bool],
        panicked: bool = False,
    ) -> Awaitable[None] | None:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "... disconnected!"
        )
        self.status = BootStatus.OFFLINE
        if not boot_future.done():
            boot_future.set_result(False)
        if not exit_future.done():
            exit_future.set_result(not panicked)
        return None

    def _on_healthcheck_passed(self, message: OscMessage) -> Awaitable[None] | None:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "healthcheck: passed"
        )
        self.attempts = 0
        return None

    def _remove_callback(self, callback: OscCallback) -> None:
        def delete(pattern, original_callback_map) -> None:
            key = pattern.pop(0)
            if key not in original_callback_map:
                return
            callbacks, callback_map = original_callback_map[key]
            if pattern:
                delete(pattern, callback_map)
            if callback in callbacks:
                callbacks.remove(callback)
            if not callbacks and not callback_map:
                original_callback_map.pop(key)

        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            delete(list(pattern), self.callbacks)

    def _register(
        self,
        pattern,
        procedure,
        *,
        failure_pattern=None,
        once: bool = False,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> OscCallback:
        if isinstance(pattern, (str, int, float)):
            pattern = [pattern]
        if isinstance(failure_pattern, (str, int, float)):
            failure_pattern = [failure_pattern]
        if not callable(procedure):
            raise ValueError(procedure)
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"registering pattern: {pattern!r}"
        )
        return OscCallback(
            protocol=self,
            pattern=tuple(pattern),
            failure_pattern=failure_pattern,
            procedure=procedure,
            once=bool(once),
            args=args,
            kwargs=kwargs,
        )

    def _send(self, raw_message: SequenceABC | SupportsOsc | str) -> bytes:
        if self.status not in (BootStatus.BOOTING, BootStatus.ONLINE):
            raise OscProtocolOffline
        if not isinstance(raw_message, (str, SequenceABC, SupportsOsc)):
            raise ValueError(raw_message)
        message: OscBundle | OscMessage
        if isinstance(raw_message, str):
            message = OscMessage(raw_message)
        elif isinstance(raw_message, SequenceABC):
            message = OscMessage(*raw_message)
        else:
            message = raw_message.to_osc()
        osc_out_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] {message!r}"
        )
        for capture in self.captures:
            capture.add_entry(
                timestamp=time.time(),
                label="S",
                message=message,
                raw_message=raw_message,
            )
        datagram = message.to_datagram()
        udp_out_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] {datagram!r}"
        )
        return datagram

    def _setup(
        self, ip_address: str, port: int, healthcheck: HealthCheck | None
    ) -> None:
        self.status = BootStatus.BOOTING
        self.ip_address = ip_address
        self.port = port
        self.healthcheck = healthcheck
        if self.healthcheck:
            self.healthcheck_osc_callback = self.register(
                pattern=self.healthcheck.response_pattern,
                procedure=self._on_healthcheck_passed,
            )

    def _validate_receive(
        self, datagram
    ) -> Generator[tuple[OscCallback, OscMessage], None, None]:
        udp_in_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] {datagram}"
        )
        try:
            message = OscMessage.from_datagram(datagram)
        except Exception:
            raise
        osc_in_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] {message!r}"
        )
        for capture in self.captures:
            capture.add_entry(timestamp=time.time(), label="R", message=message)
        for callback in self._match_callbacks(message):
            yield callback, message

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        raise NotImplementedError

    def capture(self) -> "Capture":
        return Capture(self)

    def disconnect(self) -> Awaitable[None] | None:
        raise NotImplementedError

    def register(
        self,
        pattern: Sequence[float | str],
        procedure: Callable[[OscMessage], Awaitable[None] | None],
        *,
        failure_pattern: Sequence[float | str] | None = None,
        once: bool = False,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> OscCallback:
        raise NotImplementedError

    def send(self, message: SequenceABC | SupportsOsc | str) -> None:
        raise NotImplementedError

    def unregister(self, callback: OscCallback) -> None:
        raise NotImplementedError


class ThreadedOscProtocol(OscProtocol):
    class Server(socketserver.UDPServer):
        osc_protocol: "ThreadedOscProtocol"

        def verify_request(self, request, client_address) -> bool:
            self.osc_protocol._process_command_queue()
            return True

        def service_actions(self) -> None:
            if cast(HealthCheck, self.osc_protocol.healthcheck).active:
                self.osc_protocol._run_healthcheck()

    class Handler(socketserver.BaseRequestHandler):
        def handle(self) -> None:
            data = self.request[0]
            for callback, message in cast(
                ThreadedOscProtocol.Server, self.server
            ).osc_protocol._validate_receive(data):
                callback.procedure(
                    message, *(callback.args or ()), **(callback.kwargs or {})
                )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: str | None = None,
        on_connect_callback: Callable | None = None,
        on_disconnect_callback: Callable | None = None,
        on_panic_callback: Callable | None = None,
    ):
        OscProtocol.__init__(
            self,
            name=name,
            on_connect_callback=on_connect_callback,
            on_disconnect_callback=on_disconnect_callback,
            on_panic_callback=on_panic_callback,
        )
        self.boot_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self.exit_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self.command_queue: Queue[tuple[Literal["add", "remove"], OscCallback]] = (
            Queue()
        )
        self.healthcheck_deadline = 0.0
        self.lock = threading.RLock()

    ### PRIVATE METHODS ###

    def _disconnect(self, panicked: bool = False) -> None:
        super()._disconnect(panicked=panicked)
        if not self.osc_server._BaseServer__shutdown_request:  # type: ignore
            # We set the shutdown request flag rather than call .shutdown()
            # because this is often being called from _inside_ the server
            # thread.
            # N.B. Can't figure out how to make Mypy play nice with this.
            self.osc_server._BaseServer__shutdown_request = True  # type: ignore
        self._on_disconnect(
            boot_future=self.boot_future,
            exit_future=self.exit_future,
            panicked=panicked,
        )

    def _on_connect(self, boot_future: FutureLike[bool]) -> None:
        super()._on_connect(boot_future=boot_future)
        if self.on_connect_callback:
            self.on_connect_callback()

    def _on_disconnect(
        self,
        *,
        boot_future: FutureLike[bool],
        exit_future: FutureLike[bool],
        panicked: bool = False,
    ) -> None:
        super()._on_disconnect(
            boot_future=boot_future,
            exit_future=exit_future,
            panicked=panicked,
        )
        if panicked and self.on_panic_callback:
            self.on_panic_callback()
        elif not panicked and self.on_disconnect_callback:
            self.on_disconnect_callback()

    def _on_healthcheck_passed(self, message: OscMessage) -> None:
        super()._on_healthcheck_passed(message)
        if self.status == BootStatus.BOOTING:
            self._on_connect(boot_future=self.boot_future)

    def _process_command_queue(self) -> None:
        while self.command_queue.qsize():
            try:
                action, callback = self.command_queue.get()
            except Empty:
                continue
            if action == "add":
                self._add_callback(callback)
            elif action == "remove":
                self._remove_callback(callback)

    def _run_healthcheck(self) -> None:
        if self.healthcheck is None:
            return
        now = time.time()
        if now < self.healthcheck_deadline:
            return
        if self.attempts > 0:
            remaining = self.healthcheck.max_attempts - self.attempts
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                f"healthcheck failed, {remaining} attempts remaining"
            )
        new_timeout = self.healthcheck.timeout * pow(
            self.healthcheck.backoff_factor, self.attempts
        )
        self.healthcheck_deadline = now + new_timeout
        self.attempts += 1
        if self.attempts <= self.healthcheck.max_attempts:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "healthcheck: checking ..."
            )
            self.send(OscMessage(*self.healthcheck.request_pattern))
            return
        self._disconnect(panicked=True)

    def _server_factory(self, ip_address, port) -> "Server":
        server = self.Server(
            (self.ip_address, self.port), self.Handler, bind_and_activate=False
        )
        server.osc_protocol = self
        return server

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        if self._activate_healthcheck():
            cast(HealthCheck, self.healthcheck).active = True

    def connect(
        self, ip_address: str, port: int, *, healthcheck: HealthCheck | None = None
    ) -> None:
        if self.status != BootStatus.OFFLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already connected!"
            )
            raise OscProtocolAlreadyConnected
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "connecting ..."
        )
        self._setup(ip_address, port, healthcheck)
        self.healthcheck_deadline = time.time()
        self.boot_future = concurrent.futures.Future()
        self.exit_future = concurrent.futures.Future()
        self.osc_server = self._server_factory(ip_address, port)
        self.osc_server_thread = threading.Thread(target=self.osc_server.serve_forever)
        self.osc_server_thread.daemon = True
        self.osc_server_thread.start()
        if not self.healthcheck:
            self._on_connect(boot_future=self.boot_future)

    def disconnect(self) -> None:
        if self.status != BootStatus.ONLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already disconnected!"
            )
            return
        self._disconnect()

    def register(
        self,
        pattern: Sequence[float | str],
        procedure: Callable[[OscMessage], Awaitable[None] | None],
        *,
        failure_pattern: Sequence[float | str] | None = None,
        once: bool = False,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> OscCallback:
        """
        Register a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(
            (
                "add",
                callback := self._register(
                    pattern,
                    procedure,
                    failure_pattern=failure_pattern,
                    once=once,
                    args=args,
                    kwargs=kwargs,
                ),
            )
        )
        return callback

    def send(self, message: SequenceABC | SupportsOsc | str) -> None:
        try:
            self.osc_server.socket.sendto(
                self._send(message),
                (self.ip_address, self.port),
            )
        except OSError:
            # print(message)
            raise

    def unregister(self, callback: OscCallback) -> None:
        """
        Unregister a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(("remove", callback))


class AsyncOscProtocol(asyncio.DatagramProtocol, OscProtocol):
    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: str | None = None,
        on_connect_callback: Callable | None = None,
        on_disconnect_callback: Callable | None = None,
        on_panic_callback: Callable | None = None,
    ) -> None:
        asyncio.DatagramProtocol.__init__(self)
        OscProtocol.__init__(
            self,
            name=name,
            on_connect_callback=on_connect_callback,
            on_disconnect_callback=on_disconnect_callback,
            on_panic_callback=on_panic_callback,
        )
        self.boot_future: asyncio.Future[bool] = asyncio.Future()
        self.exit_future: asyncio.Future[bool] = asyncio.Future()
        self.background_tasks: set[asyncio.Task] = set()
        self.healthcheck_task: asyncio.Task | None = None

    ### PRIVATE METHODS ###

    async def _disconnect(self, panicked: bool = False) -> None:
        super()._disconnect(panicked=panicked)
        self.transport.close()
        if self.healthcheck_task:
            self.healthcheck_task.cancel()
        await self._on_disconnect(
            boot_future=self.boot_future,
            exit_future=self.exit_future,
            panicked=panicked,
        )

    async def _on_connect(self, *, boot_future: FutureLike[bool]) -> None:
        super()._on_connect(boot_future=self.boot_future)
        if self.on_connect_callback:
            self.on_connect_callback()

    async def _on_disconnect(
        self,
        *,
        boot_future: FutureLike[bool],
        exit_future: FutureLike[bool],
        panicked: bool = False,
    ) -> None:
        super()._on_disconnect(
            boot_future=boot_future,
            exit_future=exit_future,
            panicked=panicked,
        )
        if panicked and self.on_panic_callback:
            self.on_panic_callback()
        elif not panicked and self.on_disconnect_callback:
            self.on_disconnect_callback()

    async def _on_healthcheck_passed(self, message: OscMessage) -> None:
        super()._on_healthcheck_passed(message)
        if self.status == BootStatus.BOOTING:
            await self._on_connect(boot_future=self.boot_future)

    async def _run_healthcheck(self) -> None:
        if self.healthcheck is None:
            return
        while self.status in (BootStatus.BOOTING, BootStatus.ONLINE):
            if self.attempts >= self.healthcheck.max_attempts:
                await self._disconnect(panicked=True)
                return
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "healthcheck: checking ..."
            )
            self.send(OscMessage(*self.healthcheck.request_pattern))
            sleep_time = self.healthcheck.timeout * pow(
                self.healthcheck.backoff_factor, self.attempts
            )
            self.attempts += 1
            await asyncio.sleep(sleep_time)

    ### OVERRIDES ###

    def connection_made(self, transport) -> None:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "connection made!"
        )
        self.transport = transport

    def connection_lost(self, exc) -> None:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "connection lost!"
        )

    def datagram_received(self, data, addr) -> None:
        loop = asyncio.get_running_loop()
        for callback, message in self._validate_receive(data):
            if asyncio.iscoroutine(
                result := callback.procedure(
                    message, *(callback.args or ()), **(callback.kwargs or {})
                )
            ):
                self.background_tasks.add(task := loop.create_task(result))
                task.add_done_callback(self.background_tasks.discard)

    def error_received(self, exc) -> None:
        osc_out_logger.warning(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"errored: {exc}"
        )

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        if self._activate_healthcheck():
            self.healthcheck_task = asyncio.get_running_loop().create_task(
                self._run_healthcheck()
            )

    async def connect(
        self, ip_address: str, port: int, *, healthcheck: HealthCheck | None = None
    ):
        if self.status != BootStatus.OFFLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already connected!"
            )
            raise OscProtocolAlreadyConnected
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "connecting ..."
        )
        self._setup(ip_address, port, healthcheck)
        loop = asyncio.get_running_loop()
        self.boot_future = loop.create_future()
        self.exit_future = loop.create_future()
        _, protocol = await loop.create_datagram_endpoint(
            lambda: self, remote_addr=(ip_address, port)
        )
        if self.healthcheck and self.healthcheck.active:
            self.healthcheck_task = loop.create_task(self._run_healthcheck())
        elif not self.healthcheck:
            await self._on_connect(boot_future=self.boot_future)

    async def disconnect(self) -> None:
        if self.status != BootStatus.ONLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already disconnected!"
            )
            return
        await self._disconnect()

    def register(
        self,
        pattern: Sequence[float | str],
        procedure: Callable[[OscMessage], Awaitable[None] | None],
        *,
        failure_pattern: Sequence[float | str] | None = None,
        once: bool = False,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> OscCallback:
        """
        Register a callback.
        """
        self._add_callback(
            callback := self._register(
                pattern,
                procedure,
                failure_pattern=failure_pattern,
                once=once,
                args=args,
                kwargs=kwargs,
            )
        )
        return callback

    def send(self, message: SequenceABC | SupportsOsc | str) -> None:
        self.transport.sendto(self._send(message))

    def unregister(self, callback: OscCallback) -> None:
        self._remove_callback(callback)



from uuid import UUID

from uqbar.objects import get_vars, new

from supriya.patterns.events import CompositeEvent, Event


def sanitize_id(id_: UUID, cache: dict[UUID, UUID]) -> UUID:
    if id_ not in cache:
        cache[id_] = UUID(int=len(cache))
    return cache[id_]


def sanitize_event(event: Event, cache: dict[UUID, UUID]) -> Event:
    if isinstance(event, CompositeEvent):
        return new(event, events=[sanitize_event(x, cache) for x in event.events])
    sanitize_data = {}
    args, _, kwargs = get_vars(event)
    for key, value in args.items():
        if isinstance(value, UUID):
            value = sanitize_id(value, cache)
        sanitize_data[key] = value
    for key, value in sorted(kwargs.items()):
        if isinstance(value, UUID):
            value = sanitize_id(value, cache)
        sanitize_data[key] = value
    return type(event)(**sanitize_data)


def sanitize(exprs):
    cache = {}
    sanitized = []
    for expr in exprs:
        if isinstance(expr, Event):
            sanitized.append(sanitize_event(expr, cache))
        else:
            sanitized.append(expr)
    return sanitized


def run_pattern_test(pattern, expected, is_infinite, stop_at):
    assert pattern.is_infinite == is_infinite
    iterator = iter(pattern)
    actual = []
    ceased = True
    for iteration in range(1000):
        try:
            if stop_at == iteration:
                expr = iterator.send(True)
            else:
                expr = next(iterator)
            actual.append(expr)
        except StopIteration:
            break
    else:
        ceased = False
    if is_infinite:
        assert not ceased
        sanitized_actual = sanitize(actual[: len(expected)])
    else:
        sanitized_actual = sanitize(actual)
    assert sanitized_actual == expected, sanitized_actual



import pytest
from uqbar.strings import normalize

from supriya.ugens import Rand, SinOsc, Sum4
from supriya.ugens.core import UGenRecursiveInput


@pytest.mark.parametrize(
    "input_one, input_two, input_three, input_four, expected",
    [
        (
            1.0,
            SinOsc.kr(),
            SinOsc.ar(),
            Rand.ir(),
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.ar:
                        frequency: 440.0
                        phase: 0.0
                -   SinOsc.kr:
                        frequency: 440.0
                        phase: 0.0
                -   Rand.ir:
                        minimum: 0.0
                        maximum: 1.0
                -   Sum4.ar:
                        input_one: SinOsc.ar[0]
                        input_two: SinOsc.kr[0]
                        input_three: 1.0
                        input_four: Rand.ir[0]
            """,
        ),
        (
            0,
            SinOsc.ar(),
            Rand.ir(),
            1.0,
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.ar:
                        frequency: 440.0
                        phase: 0.0
                -   Rand.ir:
                        minimum: 0.0
                        maximum: 1.0
                -   Sum3.ar:
                        input_one: SinOsc.ar[0]
                        input_two: Rand.ir[0]
                        input_three: 1.0
            """,
        ),
        (
            SinOsc.kr(),
            0,
            Rand.ir(),
            1.0,
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.kr:
                        frequency: 440.0
                        phase: 0.0
                -   Rand.ir:
                        minimum: 0.0
                        maximum: 1.0
                -   Sum3.kr:
                        input_one: SinOsc.kr[0]
                        input_two: Rand.ir[0]
                        input_three: 1.0
            """,
        ),
        (
            SinOsc.kr(),
            SinOsc.ar(),
            0,
            1.0,
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.ar:
                        frequency: 440.0
                        phase: 0.0
                -   SinOsc.kr:
                        frequency: 440.0
                        phase: 0.0
                -   Sum3.ar:
                        input_one: SinOsc.ar[0]
                        input_two: SinOsc.kr[0]
                        input_three: 1.0
            """,
        ),
        (
            SinOsc.kr(),
            SinOsc.ar(),
            Rand.ir(),
            0,
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.ar:
                        frequency: 440.0
                        phase: 0.0
                -   SinOsc.kr:
                        frequency: 440.0
                        phase: 0.0
                -   Rand.ir:
                        minimum: 0.0
                        maximum: 1.0
                -   Sum3.ar:
                        input_one: SinOsc.ar[0]
                        input_two: SinOsc.kr[0]
                        input_three: Rand.ir[0]
            """,
        ),
        (
            0,
            SinOsc.ar(),
            Rand.ir(),
            0,
            """
            synthdef:
                name: ...
                ugens:
                -   SinOsc.ar:
                        frequency: 440.0
                        phase: 0.0
                -   Rand.ir:
                        minimum: 0.0
                        maximum: 1.0
                -   BinaryOpUGen(ADDITION).ar:
                        left: SinOsc.ar[0]
                        right: Rand.ir[0]
            """,
        ),
    ],
)
def test_Sum4(
    input_one: UGenRecursiveInput,
    input_two: UGenRecursiveInput,
    input_three: UGenRecursiveInput,
    input_four: UGenRecursiveInput,
    expected: str,
) -> None:
    result = Sum4.new(
        input_one=input_one,
        input_two=input_two,
        input_three=input_three,
        input_four=input_four,
    )
    assert normalize(str(result)) == normalize(expected)



from typing import Literal

from ..enums import CalculationRate, DoneAction, EnvelopeShape, ParameterRate
from . import (
    FFT,
    K2A,
    LPF,
    BufDur,
    BufRd,
    BufSamples,
    Envelope,
    EnvGen,
    In,
    InFeedback,
    LFSaw,
    Linen,
    LocalBuf,
    Mix,
    MulAdd,
    OffsetOut,
    Out,
    Pan2,
    Parameter,
    PV_Div,
    PV_MagSmear,
    Rand,
    ScopeOut,
    ScopeOut2,
    SynthDef,
    SynthDefBuilder,
    VarSaw,
    XLine,
)


def _build_default_synthdef() -> SynthDef:
    with SynthDefBuilder(
        amplitude=0.1,
        frequency=440,
        gate=1,
        out=Parameter(rate=ParameterRate.SCALAR, value=0),
        pan=0.5,
    ) as builder:
        low_pass = LPF.ar(
            source=Mix.new(
                VarSaw.ar(
                    frequency=builder["frequency"]
                    + (
                        0,
                        Rand.ir(minimum=-0.4, maximum=0.0),
                        Rand.ir(minimum=0.0, maximum=0.4),
                    ),
                    width=0.3,
                )
            )
            * 0.3,
            frequency=XLine.kr(
                start=Rand.ir(minimum=4000, maximum=5000),
                stop=Rand.ir(minimum=2500, maximum=3200),
            ),
        )
        linen = Linen.kr(
            attack_time=0.01,
            done_action=DoneAction.FREE_SYNTH,
            gate=builder["gate"],
            release_time=0.3,
            sustain_level=0.7,
        )
        pan = Pan2.ar(
            source=low_pass * linen * builder["amplitude"], position=builder["pan"]
        )
        OffsetOut.ar(bus=builder["out"], source=pan)
    return builder.build(name="supriya:default")


def _build_link_audio_synthdef(channel_count: int) -> SynthDef:
    with SynthDefBuilder(
        out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    ) as builder:
        start_value = builder["fade_time"] <= 0
        envelope = Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=[EnvelopeShape.SINE],
            release_node=1,
        )
        envgen = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        source = InFeedback.ar(bus=builder["in_"], channel_count=channel_count)
        Out.ar(bus=builder["out"], source=source * envgen)
    return builder.build(name=f"supriya:link-ar:{channel_count}")


def _build_link_control_synthdef(channel_count: int) -> SynthDef:
    with SynthDefBuilder(
        out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    ) as builder:
        start_value = builder["fade_time"] <= 0
        envelope = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=Envelope(
                amplitudes=[start_value, 1.0, 0.0],
                durations=[1.0, 1.0],
                curves=[EnvelopeShape.LINEAR],
                release_node=1,
            ),
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        input_ = In.kr(bus=builder["in_"], channel_count=channel_count)
        Out.kr(bus=builder["out"], source=input_ * envelope)
    return builder.build(name=f"supriya:link-kr:{channel_count}")


def _build_amplitude_scope_synthdef(
    channel_count: int, rate: CalculationRate
) -> SynthDef:
    with SynthDefBuilder(
        in_=0,
        max_frames=4096,
        scope_frames=4096,
        scope_id=0,
    ) as builder:
        if rate is CalculationRate.AUDIO:
            source = In.ar(
                bus=builder["in_"],
                channel_count=channel_count,
            )
        elif rate is CalculationRate.CONTROL:
            source = K2A.ar(
                source=In.kr(
                    bus=builder["in_"],
                    channel_count=channel_count,
                ),
            )
        else:
            raise ValueError(rate)
        ScopeOut2.ar(
            max_frames=builder["max_frames"],
            scope_frames=builder["scope_frames"],
            scope_id=builder["scope_id"],
            source=source,
        )
    return builder.build(name=f"supriya:amp-scope-{rate.token}:{channel_count}")


def _build_frequency_scope_synthdef(
    channel_mode: Literal["mono", "stereo"] = "mono",
    frequency_mode: Literal["linear", "logarithmic"] = "linear",
    use_shared_memory: bool = True,
) -> SynthDef:
    with SynthDefBuilder(
        in_=0,
        fft_buffer_size=Parameter(value=2048, rate="SCALAR"),
        scope_id=Parameter(value=0, rate="SCALAR"),
        rate=Parameter(value=4, rate="SCALAR"),
    ) as builder:
        phase = 1 - (builder["rate"] * builder["fft_buffer_size"].reciprocal())

        fft_buffer_id = LocalBuf.ir(frame_count=builder["fft_buffer_size"])

        sample_count = BufSamples.ir(buffer_id=fft_buffer_id) * 0.5

        source = In.ar(channel_count=1, bus=builder["in_"])

        pv_chain = FFT.kr(
            buffer_id=fft_buffer_id,
            hop=0.75,
            source=source,
            window_type=1,
        )

        if frequency_mode == "stereo":
            source_2 = In.ar(
                bus=builder.add_parameter(name="in2", value=0),
                channel_count=1,
            )
            fft_buffer_id_2 = LocalBuf.ir(frame_count=builder["fft_buffer_size"])
            pv_chain_2 = FFT.kr(
                buffer_id=fft_buffer_id_2,
                hop=0.75,
                source=source_2,
                window_type=1,
            )
            pv_chain = PV_Div(pv_chain_a=pv_chain_2, pv_chain_b=pv_chain)

        pv_chain = PV_MagSmear.kr(bins=1, pv_chain=pv_chain)

        if frequency_mode == "linear":
            phasor = MulAdd.new(
                source=LFSaw.ar(
                    frequency=builder["rate"] / BufDur.ir(buffer_id=fft_buffer_id),
                    initial_phase=phase,
                ),
                multiplier=sample_count,
                addend=sample_count,
            )
        else:
            phasor = (
                sample_count
                ** (
                    MulAdd.new(
                        source=LFSaw.ar(
                            frequency=builder["rate"]
                            / BufDur.ir(buffer_id=fft_buffer_id),
                            initial_phase=phase,
                        ),
                        multiplier=0.5,
                        addend=0.5,
                    )
                )
            ) * 2
        phasor = phasor.round(2)

        scope_source = (
            BufRd.ar(
                buffer_id=fft_buffer_id,
                channel_count=1,
                interpolation=1,
                loop=1,
                phase=phasor,
            )
            / builder["fft_buffer_size"]
        ).amplitude_to_db()

        if not use_shared_memory:
            ScopeOut.ar(
                buffer_id=builder["scope_id"],
                source=scope_source,
            )
        else:
            ScopeOut2.ar(
                scope_id=builder["scope_id"],
                source=scope_source,
                max_frames=builder["fft_buffer_size"] / builder["rate"],
            )
    name = "supriya:freq-scope"
    name += "-lin" if frequency_mode == "linear" else "-log"
    name += "-shm" if use_shared_memory else ""
    name += ":2" if channel_mode == "stereo" else ":1"
    return builder.build(name=name)


# default synthdef
default = _build_default_synthdef()

# audio-rate link synthdefs
system_link_audio_1 = _build_link_audio_synthdef(1)
system_link_audio_2 = _build_link_audio_synthdef(2)
system_link_audio_3 = _build_link_audio_synthdef(3)
system_link_audio_4 = _build_link_audio_synthdef(4)
system_link_audio_5 = _build_link_audio_synthdef(5)
system_link_audio_6 = _build_link_audio_synthdef(6)
system_link_audio_7 = _build_link_audio_synthdef(7)
system_link_audio_8 = _build_link_audio_synthdef(8)
system_link_audio_9 = _build_link_audio_synthdef(9)
system_link_audio_10 = _build_link_audio_synthdef(10)
system_link_audio_11 = _build_link_audio_synthdef(11)
system_link_audio_12 = _build_link_audio_synthdef(12)
system_link_audio_13 = _build_link_audio_synthdef(13)
system_link_audio_14 = _build_link_audio_synthdef(14)
system_link_audio_15 = _build_link_audio_synthdef(15)
system_link_audio_16 = _build_link_audio_synthdef(16)

# control-rate link synthdefs
system_link_control_1 = _build_link_control_synthdef(1)
system_link_control_2 = _build_link_control_synthdef(2)
system_link_control_3 = _build_link_control_synthdef(3)
system_link_control_4 = _build_link_control_synthdef(4)
system_link_control_5 = _build_link_control_synthdef(5)
system_link_control_6 = _build_link_control_synthdef(6)
system_link_control_7 = _build_link_control_synthdef(7)
system_link_control_8 = _build_link_control_synthdef(8)
system_link_control_9 = _build_link_control_synthdef(9)
system_link_control_10 = _build_link_control_synthdef(10)
system_link_control_11 = _build_link_control_synthdef(11)
system_link_control_12 = _build_link_control_synthdef(12)
system_link_control_13 = _build_link_control_synthdef(13)
system_link_control_14 = _build_link_control_synthdef(14)
system_link_control_15 = _build_link_control_synthdef(15)
system_link_control_16 = _build_link_control_synthdef(16)

# audio-rate stethoscope synthdefs
amplitude_scope_audio_1 = _build_amplitude_scope_synthdef(1, CalculationRate.AUDIO)
amplitude_scope_audio_2 = _build_amplitude_scope_synthdef(2, CalculationRate.AUDIO)
amplitude_scope_audio_3 = _build_amplitude_scope_synthdef(3, CalculationRate.AUDIO)
amplitude_scope_audio_4 = _build_amplitude_scope_synthdef(4, CalculationRate.AUDIO)
amplitude_scope_audio_5 = _build_amplitude_scope_synthdef(5, CalculationRate.AUDIO)
amplitude_scope_audio_6 = _build_amplitude_scope_synthdef(6, CalculationRate.AUDIO)
amplitude_scope_audio_7 = _build_amplitude_scope_synthdef(7, CalculationRate.AUDIO)
amplitude_scope_audio_8 = _build_amplitude_scope_synthdef(8, CalculationRate.AUDIO)
amplitude_scope_audio_9 = _build_amplitude_scope_synthdef(9, CalculationRate.AUDIO)
amplitude_scope_audio_10 = _build_amplitude_scope_synthdef(10, CalculationRate.AUDIO)
amplitude_scope_audio_11 = _build_amplitude_scope_synthdef(11, CalculationRate.AUDIO)
amplitude_scope_audio_12 = _build_amplitude_scope_synthdef(12, CalculationRate.AUDIO)
amplitude_scope_audio_13 = _build_amplitude_scope_synthdef(13, CalculationRate.AUDIO)
amplitude_scope_audio_14 = _build_amplitude_scope_synthdef(14, CalculationRate.AUDIO)
amplitude_scope_audio_15 = _build_amplitude_scope_synthdef(15, CalculationRate.AUDIO)
amplitude_scope_audio_15 = _build_amplitude_scope_synthdef(16, CalculationRate.AUDIO)

# control-rate amplitude_scope synthdefs
amplitude_scope_control_1 = _build_amplitude_scope_synthdef(1, CalculationRate.CONTROL)
amplitude_scope_control_2 = _build_amplitude_scope_synthdef(2, CalculationRate.CONTROL)
amplitude_scope_control_3 = _build_amplitude_scope_synthdef(3, CalculationRate.CONTROL)
amplitude_scope_control_4 = _build_amplitude_scope_synthdef(4, CalculationRate.CONTROL)
amplitude_scope_control_5 = _build_amplitude_scope_synthdef(5, CalculationRate.CONTROL)
amplitude_scope_control_6 = _build_amplitude_scope_synthdef(6, CalculationRate.CONTROL)
amplitude_scope_control_7 = _build_amplitude_scope_synthdef(7, CalculationRate.CONTROL)
amplitude_scope_control_8 = _build_amplitude_scope_synthdef(8, CalculationRate.CONTROL)
amplitude_scope_control_9 = _build_amplitude_scope_synthdef(9, CalculationRate.CONTROL)
amplitude_scope_control_10 = _build_amplitude_scope_synthdef(
    10, CalculationRate.CONTROL
)
amplitude_scope_control_11 = _build_amplitude_scope_synthdef(
    11, CalculationRate.CONTROL
)
amplitude_scope_control_12 = _build_amplitude_scope_synthdef(
    12, CalculationRate.CONTROL
)
amplitude_scope_control_13 = _build_amplitude_scope_synthdef(
    13, CalculationRate.CONTROL
)
amplitude_scope_control_14 = _build_amplitude_scope_synthdef(
    14, CalculationRate.CONTROL
)
amplitude_scope_control_15 = _build_amplitude_scope_synthdef(
    15, CalculationRate.CONTROL
)
amplitude_scope_control_15 = _build_amplitude_scope_synthdef(
    16, CalculationRate.CONTROL
)

# frequency scope synthdefs
frequency_scope_lin_1 = _build_frequency_scope_synthdef(
    channel_mode="mono", frequency_mode="linear", use_shared_memory=False
)
frequency_scope_lin_2 = _build_frequency_scope_synthdef(
    channel_mode="stereo", frequency_mode="linear", use_shared_memory=False
)
frequency_scope_lin_shm_1 = _build_frequency_scope_synthdef(
    channel_mode="mono", frequency_mode="linear", use_shared_memory=True
)
frequency_scope_lin_shm_2 = _build_frequency_scope_synthdef(
    channel_mode="stereo", frequency_mode="linear", use_shared_memory=True
)
frequency_scope_log_1 = _build_frequency_scope_synthdef(
    channel_mode="mono", frequency_mode="logarithmic", use_shared_memory=False
)
frequency_scope_log_2 = _build_frequency_scope_synthdef(
    channel_mode="stereo", frequency_mode="logarithmic", use_shared_memory=False
)
frequency_scope_log_shm_1 = _build_frequency_scope_synthdef(
    channel_mode="mono", frequency_mode="logarithmic", use_shared_memory=True
)
frequency_scope_log_shm_2 = _build_frequency_scope_synthdef(
    channel_mode="stereo", frequency_mode="logarithmic", use_shared_memory=True
)

AMPLITUDE_SCOPE_SYNTHDEFS: dict[str, SynthDef] = {
    synthdef.effective_name: synthdef
    for synthdef in [
        amplitude_scope_audio_1,
        amplitude_scope_audio_2,
        amplitude_scope_audio_3,
        amplitude_scope_audio_4,
        amplitude_scope_audio_5,
        amplitude_scope_audio_6,
        amplitude_scope_audio_7,
        amplitude_scope_audio_8,
        amplitude_scope_audio_9,
        amplitude_scope_audio_10,
        amplitude_scope_audio_11,
        amplitude_scope_audio_12,
        amplitude_scope_audio_13,
        amplitude_scope_audio_14,
        amplitude_scope_audio_15,
        amplitude_scope_audio_15,
        amplitude_scope_control_1,
        amplitude_scope_control_2,
        amplitude_scope_control_3,
        amplitude_scope_control_4,
        amplitude_scope_control_5,
        amplitude_scope_control_6,
        amplitude_scope_control_7,
        amplitude_scope_control_8,
        amplitude_scope_control_9,
        amplitude_scope_control_10,
        amplitude_scope_control_11,
        amplitude_scope_control_12,
        amplitude_scope_control_13,
        amplitude_scope_control_14,
        amplitude_scope_control_15,
        amplitude_scope_control_15,
    ]
}

FREQUENCY_SCOPE_SYNTHDEFS: dict[str, SynthDef] = {
    synthdef.effective_name: synthdef
    for synthdef in [
        frequency_scope_lin_1,
        frequency_scope_lin_2,
        frequency_scope_log_1,
        frequency_scope_log_2,
        frequency_scope_lin_shm_1,
        frequency_scope_lin_shm_2,
        frequency_scope_log_shm_1,
        frequency_scope_log_shm_2,
    ]
}

SYSTEM_SYNTHDEFS: dict[str, SynthDef] = {
    synthdef.effective_name: synthdef
    for synthdef in [
        system_link_audio_1,
        system_link_audio_2,
        system_link_audio_3,
        system_link_audio_4,
        system_link_audio_5,
        system_link_audio_6,
        system_link_audio_7,
        system_link_audio_8,
        system_link_audio_9,
        system_link_audio_10,
        system_link_audio_11,
        system_link_audio_12,
        system_link_audio_13,
        system_link_audio_14,
        system_link_audio_15,
        system_link_audio_16,
        system_link_control_1,
        system_link_control_2,
        system_link_control_3,
        system_link_control_4,
        system_link_control_5,
        system_link_control_6,
        system_link_control_7,
        system_link_control_8,
        system_link_control_9,
        system_link_control_10,
        system_link_control_11,
        system_link_control_12,
        system_link_control_13,
        system_link_control_14,
        system_link_control_15,
        system_link_control_16,
    ]
}

__all__ = [
    "SPECTROSCOPE_SYNTHDEFS",
    "STETHOSCOPE_SYNTHDEFS",
    "SYSTEM_SYNTHDEFS",
    "default",
]



from unittest.mock import Mock, call
from uuid import UUID, uuid4

import pytest
from pytest_mock import MockerFixture

from supriya import AddAction, CalculationRate, Server, Synth, default
from supriya.contexts import BusGroup, ContextObject
from supriya.patterns.events import BusAllocateEvent, Event, NoteEvent, Priority

id_ = uuid4()
bus_id = uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [
        (
            NoteEvent(id_),
            0.0,
            [
                (0.0, Priority.START, NoteEvent((id_, 0))),
                (1.0, Priority.STOP, NoteEvent((id_, 0))),
            ],
        ),
        (
            NoteEvent(id_, frequency=[440, 550]),
            0.0,
            [
                (0.0, Priority.START, NoteEvent((id_, 0), frequency=440)),
                (0.0, Priority.START, NoteEvent((id_, 1), frequency=550)),
                (1.0, Priority.STOP, NoteEvent((id_, 0), frequency=440)),
                (1.0, Priority.STOP, NoteEvent((id_, 1), frequency=550)),
            ],
        ),
        (
            NoteEvent(id_, frequency=[440, 550], amplitude=[0.5, 0.75, 1.0]),
            2.5,
            [
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 0), amplitude=0.5, frequency=440),
                ),
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 1), amplitude=0.75, frequency=550),
                ),
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 2), amplitude=1.0, frequency=440),
                ),
                (3.5, Priority.STOP, NoteEvent((id_, 0), amplitude=0.5, frequency=440)),
                (
                    3.5,
                    Priority.STOP,
                    NoteEvent((id_, 1), amplitude=0.75, frequency=550),
                ),
                (3.5, Priority.STOP, NoteEvent((id_, 2), amplitude=1.0, frequency=440)),
            ],
        ),
    ],
)
def test_expand(
    event: Event, offset: float, expected: list[tuple[float, Priority, Event]]
) -> None:
    actual = event.expand(offset)
    assert actual == expected


def test_perform(mocker: MockerFixture) -> None:
    context = Server().boot()
    spy = Mock(wraps=context)
    mocker.patch.object(context, "send")
    proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject] = {}
    notes_mapping: dict[UUID | tuple[UUID, int], float] = {}
    # 0: Allocate a bus for later
    bus_event = BusAllocateEvent(bus_id, calculation_rate="control")
    with context.at():
        bus_event.perform(
            context,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    # A: Allocate
    event_one = NoteEvent(id_)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        ),
    }
    assert notes_mapping == {id_: 1.0}
    assert spy.mock_calls == [
        call.add_synth(
            add_action=AddAction.ADD_TO_HEAD,
            permanent=False,
            synthdef=default,
            target_node=None,
        )
    ]
    # B: Already allocated, so update settings
    event_two = NoteEvent(id_, amplitude=bus_id, frequency=550)
    spy = Mock(wraps=context)
    with context.at():
        event_two.perform(
            spy,
            proxy_mapping,
            current_offset=1.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            count=1,
            calculation_rate=CalculationRate.CONTROL,
        ),
    }
    assert notes_mapping == {id_: 2.0}
    assert spy.mock_calls == [
        call.set_node(
            Synth(context=context, id_=1000, synthdef=default),
            # This doesn't make a lot of sense in real world usage, but we do
            # want to validate that the bus ID gets used
            amplitude=proxy_mapping[bus_id].id_,
            frequency=550,
        )
    ]
    # C: Free, but stop time doesn't match: no-op
    spy = Mock(wraps=context)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=1.0,
            notes_mapping=notes_mapping,
            priority=Priority.STOP,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        ),
    }
    assert notes_mapping == {id_: 2.0}
    assert spy.mock_calls == []
    # D: Free, and stop time does match
    spy = Mock(wraps=context)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=2.0,
            notes_mapping=notes_mapping,
            priority=Priority.STOP,
        )
    assert proxy_mapping == {
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        )
    }
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.free_node(Synth(context=context, id_=1000, synthdef=default))
    ]



import asyncio
import contextlib
import difflib
import logging
import textwrap
import time
import warnings
from typing import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from pytest_lazy_fixtures import lf

import supriya


async def get(x):
    # if x is awaitable, await it
    if asyncio.iscoroutine(x):
        return await x
    # otherwise just return it
    return x


@pytest.fixture
def server() -> Generator[supriya.Server, None, None]:
    # setup
    server = supriya.Server(name="test-server").boot()
    # yield to test
    yield server
    # teardown, even if the test fails
    server.quit()


@pytest_asyncio.fixture
async def async_server() -> AsyncGenerator[supriya.AsyncServer, None]:
    # setup
    server = await supriya.AsyncServer(name="test-server").boot()
    # yield to test
    yield server
    # teardown, even if the test fails
    await server.quit()


@pytest_asyncio.fixture(params=[supriya.AsyncServer, supriya.Server])
async def context(
    request,
) -> AsyncGenerator[supriya.AsyncServer | supriya.Server, None]:
    # this fixture will cause tests using it to run twice:
    # the context class is first AsyncServer, then Server
    context_class = request.param
    # instantiation looks the same
    context = context_class(name="test-server")
    # booting might need to be awaited
    await get(context.boot())
    # yield the context
    yield context
    # quitting might need to be awaited
    await get(context.quit())


@pytest.mark.asyncio
async def test_async(async_server: supriya.AsyncServer) -> None:
    # operation
    actual_tree = str(await async_server.query_tree())
    # validation
    expected_tree = textwrap.dedent(
        """\
        NODE TREE 0 group
            1 group
        """
    )
    assert actual_tree == expected_tree


@pytest.mark.asyncio
async def test_async_and_sync(context: supriya.AsyncServer | supriya.Server) -> None:
    # operation
    actual_tree = str(await get(context.query_tree()))
    # validate they're the same
    expected_tree = textwrap.dedent(
        """\
        NODE TREE 0 group
            1 group
        """
    )
    assert actual_tree == expected_tree


def test_basic() -> None:
    # setup
    server = supriya.Server().boot()
    # operation
    actual_tree = str(server.query_tree())
    # validation
    expected_tree = textwrap.dedent(
        """\
        NODE TREE 0 group
            1 group
        """
    )
    assert actual_tree == expected_tree
    # teardown (but might not happen if the assert fails)
    server.quit()


def test_exceptions(server: supriya.Server) -> None:
    # if the server is already online,
    # booting again will throw an exception
    with pytest.raises(supriya.exceptions.ServerOnline) as exception_info:
        # the test fails if the exception doesn't raise
        server.boot()
    # and we can assert even more about the raised exception
    # e.g. what the custom message for the exception was
    assert "Server already online!" in str(exception_info.value)


def test_fixtures(
    # note: the server argument name matches the server fixture name exactly
    server: supriya.Server,
) -> None:
    # operation
    actual_tree = str(server.query_tree())
    # validation
    expected_tree = textwrap.dedent(
        """\
        NODE TREE 0 group
            1 group
        """
    )
    assert actual_tree == expected_tree


def test_logging(caplog: pytest.LogCaptureFixture, server: supriya.Server) -> None:
    # capture logs from the supriya.osc.out logger
    # that are at or higher than logging.DEBUG
    caplog.set_level(logging.DEBUG, logger="supriya.osc.out")
    # do something that should emit some osc logging
    server.add_group()
    # validate that the logs are what we expect...
    # ``caplog.record_tuples`` is a list of triples: logger, level, message
    assert caplog.record_tuples == [
        (
            "supriya.osc.out",
            logging.DEBUG,
            f"[127.0.0.1:57110/{server.name}] OscMessage('/g_new', 1000, 0, 1)",
        ),
    ]


def test_node_tree_diff(server: supriya.Server) -> None:
    # capture the tree before
    before_tree = str(server.query_tree())
    # perform an operation
    with server.at():
        for _ in range(5):
            server.add_group()
    # capture the tree after
    after_tree = str(server.query_tree())
    # build a diff of the two trees
    actual_diff = "".join(
        difflib.unified_diff(
            before_tree.splitlines(True),
            after_tree.splitlines(True),
            fromfile="before",
            tofile="after",
        )
    )
    # tidy up the expected diff
    expected_diff = textwrap.dedent(
        """\
        --- before
        +++ after
        @@ -1,2 +1,7 @@
         NODE TREE 0 group
             1 group
        +        1004 group
        +        1003 group
        +        1002 group
        +        1001 group
        +        1000 group
        """
    )
    # validate
    assert actual_diff == expected_diff


@contextlib.asynccontextmanager
async def assert_node_tree_diff(
    context: supriya.AsyncServer | supriya.Server,
    expected_diff: str,
) -> AsyncGenerator[None, None]:
    # capture the tree before
    before_tree = str(await get(context.query_tree()))
    # run the body of the context manager block
    yield
    # capture the tree after
    after_tree = str(await get(context.query_tree()))
    # build a diff of the two trees
    actual_diff = "".join(
        difflib.unified_diff(
            before_tree.splitlines(True),
            after_tree.splitlines(True),
            fromfile="before",
            tofile="after",
        )
    )
    # tidy up the expected diff so we don't need to dedent inside our tests
    expected_diff = textwrap.dedent(expected_diff)
    # validate that they match
    assert actual_diff == expected_diff


@pytest.mark.asyncio
async def test_node_tree_diff_context_manager(
    context: supriya.AsyncServer | supriya.Server,
) -> None:
    # the validation happens once we exit this context manager
    async with assert_node_tree_diff(
        context=context,
        expected_diff="""\
            --- before
            +++ after
            @@ -1,2 +1,7 @@
             NODE TREE 0 group
                 1 group
            +        1004 group
            +        1003 group
            +        1002 group
            +        1001 group
            +        1000 group
        """,
    ):
        # perform an operation
        with context.at():
            for _ in range(5):
                context.add_group()


def test_osc_transcript(server: supriya.Server) -> None:
    # sniff OSC messages going to and coming from the server
    with server.osc_protocol.capture() as transcript:
        # perform an action that will emit OSC messages
        server.add_group()
    # validate that sent messages match what we expect
    assert [entry.message for entry in transcript.filtered(received=False)] == [
        supriya.OscMessage("/g_new", 1000, 0, 1),
    ]


@pytest.mark.parametrize(
    "add_action, expected_diff",
    [
        (
            supriya.AddAction.ADD_TO_HEAD,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
                     1000 group
            +            1002 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_TO_TAIL,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +            1002 group
            """,
        ),
        (
            supriya.AddAction.ADD_BEFORE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
            +        1002 group
                     1000 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_AFTER,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +        1002 group
            """,
        ),
        (
            supriya.AddAction.REPLACE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,3 @@
             NODE TREE 0 group
                 1 group
            -        1000 group
            -            1001 group
            +        1002 group
            """,
        ),
    ],
)
@pytest.mark.asyncio
async def test_parametrized(
    add_action: supriya.AddAction,
    expected_diff: str,
    server: supriya.Server,
) -> None:
    # add a group
    group = server.add_group()
    # make sure the group has a child group
    group.add_group()
    async with assert_node_tree_diff(
        context=server,
        expected_diff=expected_diff,
    ):
        # add another group relative to the first one
        # with the add action varying depending on the scenario
        group.add_group(add_action=add_action)


@pytest.mark.parametrize(
    "add_action, expected_diff",
    [
        (
            supriya.AddAction.ADD_TO_HEAD,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
                     1000 group
            +            1002 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_TO_TAIL,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +            1002 group
            """,
        ),
        (
            supriya.AddAction.ADD_BEFORE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
            +        1002 group
                     1000 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_AFTER,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +        1002 group
            """,
        ),
        (
            supriya.AddAction.REPLACE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,3 @@
             NODE TREE 0 group
                 1 group
            -        1000 group
            -            1001 group
            +        1002 group
            """,
        ),
    ],
)
@pytest.mark.asyncio
async def test_parametrized_combinatoric(
    add_action: supriya.AddAction,
    expected_diff: str,
    context: supriya.AsyncServer | supriya.Server,
) -> None:
    # add a group
    group = context.add_group()
    # make sure the group has a child group
    group.add_group()
    async with assert_node_tree_diff(
        context=context,
        expected_diff=expected_diff,
    ):
        # add another group relative to the first one
        # with the add action varying depending on the scenario
        group.add_group(add_action=add_action)


@pytest.mark.parametrize(
    "add_action, expected_diff",
    [
        (
            supriya.AddAction.ADD_TO_HEAD,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
                     1000 group
            +            1002 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_TO_TAIL,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +            1002 group
            """,
        ),
        (
            supriya.AddAction.ADD_BEFORE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,5 @@
             NODE TREE 0 group
                 1 group
            +        1002 group
                     1000 group
                         1001 group
            """,
        ),
        (
            supriya.AddAction.ADD_AFTER,
            """\
            --- before
            +++ after
            @@ -2,3 +2,4 @@
                 1 group
                     1000 group
                         1001 group
            +        1002 group
            """,
        ),
        (
            supriya.AddAction.REPLACE,
            """\
            --- before
            +++ after
            @@ -1,4 +1,3 @@
             NODE TREE 0 group
                 1 group
            -        1000 group
            -            1001 group
            +        1002 group
            """,
        ),
    ],
)
@pytest.mark.parametrize("context_", [lf("async_server"), lf("server")])
@pytest.mark.asyncio
async def test_parametrized_lazy_fixtures(
    add_action: supriya.AddAction,
    expected_diff: str,
    # we use a slightly different name so as to not shadow the context fixture
    # otherwise pytest will complain:
    context_: supriya.AsyncServer | supriya.Server,
) -> None:
    # add a group
    group = context_.add_group()
    # make sure the group has a child group
    group.add_group()
    async with assert_node_tree_diff(
        context=context_,
        expected_diff=expected_diff,
    ):
        # add another group relative to the first one
        # with the add action varying depending on the scenario
        group.add_group(add_action=add_action)


def test_process_transcript(server: supriya.Server) -> None:
    for _ in range(5):
        server.add_group()
    with server.process_protocol.capture() as transcript:
        # tell the server to "trace" the default group
        server.default_group.trace()
        # the trace has no end-delimiter so we don't know explicitly when to
        # stop capturing the output, so we'll just sleep and hope for the best:
        time.sleep(0.1)
    assert transcript.lines == [
        "TRACE Group 1",
        "   1004 group",
        "   1003 group",
        "   1002 group",
        "   1001 group",
        "   1000 group",
    ]


def test_try_finally() -> None:
    # setup
    server = supriya.Server().boot()
    # ugly!
    try:
        # operation
        actual_tree = str(server.query_tree())
        # validation
        expected_tree = textwrap.dedent(
            """\
            NODE TREE 0 group
                1 group
            """
        )
        assert actual_tree == expected_tree
    finally:
        # teardown (even if the assert fails)
        server.quit()


def test_warnings(server: supriya.Server) -> None:
    # let's create a client-side reference to a group that doesn't exist on the
    # server
    nonexistent_group = supriya.Group(context=server, id_=666)
    # capture warnings
    with warnings.catch_warnings(record=True) as warnings_:
        # operation: try to free a group that doesn't exist
        nonexistent_group.free()
        # wait for the server, because the warning arrives asynchronously
        server.sync()
    # validation
    assert len(warnings_) == 1
    assert str(warnings_[0].message) == "/n_free Node 666 not found"



import operator
from typing import Callable, Sequence, TypeAlias

import pytest

from supriya.patterns import BinaryOpPattern, Pattern, SequencePattern

from .conftest import run_pattern_test

PatternInput: TypeAlias = int | Sequence["PatternInput"]


@pytest.mark.parametrize(
    "stop_at, operator, input_a, input_b, expected, is_infinite",
    [
        (None, operator.add, 1, 7, [8], True),
        (None, operator.add, [1], 7, [(8,)], True),
        (None, operator.add, [1], [7], [(8,)], True),
        (None, operator.add, [[1]], [7], [((8,),)], True),
        (None, operator.add, [[[1]]], [7], [(((8,),),)], True),
        (None, operator.add, [1, 2], [7, 8], [(8, 10)], True),
        (None, operator.add, [1, 2], [7, 8, 10], [(8, 10, 11)], True),
        (None, operator.mul, SequencePattern([1, 2, 3]), 4, [4, 8, 12], False),
        (None, operator.mul, SequencePattern([1, 2, 3], None), 4, [4, 8, 12], True),
        (
            None,
            operator.add,
            SequencePattern(["a", "b", "c"]),
            SequencePattern(["1", "2", "3"]),
            ["a1", "b2", "c3"],
            False,
        ),
    ],
)
def test_pattern(
    stop_at: float | None,
    operator: Callable,
    input_a: Pattern | PatternInput,
    input_b: PatternInput,
    expected: PatternInput,
    is_infinite: bool,
) -> None:
    pattern = BinaryOpPattern(operator, input_a, input_b)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



from typing import overload

from supriya.contexts.entities import Bus as Bus
from supriya.contexts.entities import BusGroup as BusGroup

class ServerSHM:
    def __init__(self, port_number: int, bus_count: int) -> None: ...
    def describe_scope_buffer(self, index: int) -> tuple[int, int]: ...
    def read_scope_buffer(self, index: int) -> tuple[int, list[float]]: ...
    @overload
    def __getitem__(self, item: Bus | int) -> float: ...
    @overload
    def __getitem__(self, item: BusGroup | slice) -> list[float]: ...
    @overload
    def __setitem__(self, item: Bus | int, value: float) -> None: ...
    @overload
    def __setitem__(self, item: BusGroup | slice, value: list[float]) -> None: ...



import enum
import warnings
from typing import Any, Sequence, SupportsFloat
from uuid import UUID

from uqbar.objects import get_repr, get_vars, new

from ..contexts import BusGroup, Context, ContextObject, Node
from ..enums import AddAction, CalculationRate
from ..typing import AddActionLike, CalculationRateLike
from ..ugens import SynthDef, default
from ..utils import expand


def _process_synth_settings(
    proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
    **kwargs: SupportsFloat | UUID | str | Sequence[SupportsFloat | UUID | str],
) -> dict[str, float | str | Sequence[float | str]]:
    """
    Resolve UUIDs into context object ids
    """
    settings: dict[str, float | str | Sequence[float | str]] = {}
    for key, value in kwargs.items():
        if not isinstance(value, Sequence) or isinstance(value, str):
            value = [value]
        processed_values: list[float | str] = []
        for v in value:
            if isinstance(v, UUID):
                processed_values.append(float(proxy_mapping[v]))
            else:
                processed_values.append(float(v) if not isinstance(v, str) else v)
        settings[key] = (
            processed_values[0] if len(processed_values) == 1 else processed_values
        )
    return settings


class PatternWarning(Warning):
    pass


warnings.simplefilter("always", PatternWarning)


def _partition_synth_settings(
    **kwargs: float | str | Sequence[float | str],
) -> tuple[dict[str, float | Sequence[float]], dict[str, str]]:
    """
    Split synth settings into two maps suitable for set_node and map_node
    """
    set_kwargs: dict[str, float | Sequence[float]] = {}
    map_kwargs: dict[str, str] = {}
    for key, value in kwargs.items():
        if isinstance(value, float):
            set_kwargs[key] = value
        elif isinstance(value, str):
            map_kwargs[key] = value
        elif isinstance(value, Sequence):
            if all(isinstance(v, float) for v in value):
                set_kwargs[key] = [float(v) for v in value]
            else:
                warnings.warn(f"Cannot process {value}")
        # Unclear to me if we can map with a sequence of strings.
    return set_kwargs, map_kwargs


class Priority(enum.IntEnum):
    NONE = 0
    START = 1
    STOP = 2


class Event:
    def __init__(self, *, delta: float = 0.0, **kwargs) -> None:
        self.delta = delta

    def __eq__(self, expr: Any) -> bool:
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __repr__(self) -> str:
        return get_repr(self, multiline=False)

    def expand(self, offset: float) -> Sequence[tuple[float, Priority, "Event"]]:
        return [(offset, Priority.START, self)]

    def merge(self, event: "Event") -> "Event":
        _, _, kwargs = get_vars(event)
        return new(self, **kwargs)

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        pass


class StartEvent(Event):
    """
    The first event injected by a pattern player.
    """


class StopEvent(Event):
    """
    The last event injected by a pattern player.
    """


class BusAllocateEvent(Event):
    def __init__(
        self,
        id_: UUID | tuple[UUID, int],
        *,
        calculation_rate: CalculationRateLike = "audio",
        channel_count: int = 1,
        delta: float = 0.0,
        **kwargs,
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_
        self.calculation_rate = CalculationRate.from_expr(calculation_rate)
        self.channel_count = channel_count

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        proxy_mapping[self.id_] = context.add_bus_group(
            calculation_rate=self.calculation_rate, count=self.channel_count
        )


class BusFreeEvent(Event):
    def __init__(
        self, id_: UUID | tuple[UUID, int], *, delta: float = 0.0, **kwargs
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        if not isinstance(bus_group := proxy_mapping.pop(self.id_), BusGroup):
            raise RuntimeError(bus_group)
        context.free_bus_group(bus_group)


class CompositeEvent(Event):
    def __init__(self, events: list[Event], *, delta: float = 0.0, **kwargs) -> None:
        Event.__init__(self, delta=delta)
        self.events = (
            events if not kwargs else [new(event, **kwargs) for event in events]
        )

    def expand(self, offset: float) -> Sequence[tuple[float, Priority, Event]]:
        events: list[tuple[float, Priority, Event]] = []
        for event in self.events:
            events.extend(event.expand(offset))
            offset += event.delta
        return events


class NodeEvent(Event):
    def __init__(
        self,
        id_: UUID | tuple[UUID, int],
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 0.0,
        target_node: Node | UUID | None = None,
        **kwargs,
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_
        self.add_action = AddAction.from_expr(add_action)
        self.target_node = target_node

    def _resolve_target_node(
        self,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
    ) -> Node | None:
        if isinstance(self.target_node, UUID):
            if not isinstance(
                target_node_ := proxy_mapping.get(self.target_node), Node
            ):
                raise RuntimeError(target_node_)
            return target_node_
        return self.target_node


class GroupAllocateEvent(NodeEvent):
    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        proxy_mapping[self.id_] = context.add_group(
            add_action=self.add_action,
            target_node=self._resolve_target_node(
                proxy_mapping,
            ),
        )


class NodeFreeEvent(Event):
    def __init__(
        self, id_: UUID | tuple[UUID, int], *, delta: float = 0.0, **kwargs
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        if not isinstance(node := proxy_mapping.pop(self.id_), Node):
            raise RuntimeError(node)
        context.free_node(node)


class NoteEvent(NodeEvent):
    def __init__(
        self,
        id_: UUID | tuple[UUID, int],
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 1.0,
        duration: float = 1.0,
        rest: bool = False,
        synthdef: SynthDef | None = None,
        target_node: Node | UUID | None = None,
        **kwargs: SupportsFloat | UUID | Sequence[SupportsFloat | UUID],
    ) -> None:
        NodeEvent.__init__(
            self,
            add_action=add_action,
            delta=delta,
            id_=id_,
            target_node=target_node,
            **kwargs,
        )
        self.duration = duration
        self.kwargs = kwargs
        self.rest = rest
        self.synthdef = synthdef

    def expand(self, offset: float) -> Sequence[tuple[float, Priority, Event]]:
        starts, stops = [], []
        for i, proxy_mapping in enumerate(expand(self.kwargs)):
            if not isinstance(self.id_, UUID):
                raise RuntimeError("How did we get here?")
            event: Event = type(self)(
                id_=(self.id_, i),
                add_action=self.add_action,
                delta=self.delta,
                duration=self.duration,
                rest=self.rest,
                synthdef=self.synthdef,
                target_node=self.target_node,
                **proxy_mapping,
            )
            start_offset = offset
            stop_offset = offset + self.duration
            starts.append((start_offset, Priority.START, event))
            stops.append((stop_offset, Priority.STOP, event))
        return starts + stops

    def merge(self, event: Event) -> Event:
        _, _, kwargs = get_vars(event)
        return new(self, **kwargs)

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs: SupportsFloat | UUID | str | Sequence[SupportsFloat | UUID | str],
    ) -> None:
        if priority == Priority.START and not self.rest:
            # does a proxy exist?
            #    if yes, update settings
            #    if no, create proxy
            # update notes mapping with expected completion offset
            settings = _process_synth_settings(proxy_mapping, **self.kwargs)
            # add the synth
            if self.id_ not in proxy_mapping:
                proxy_mapping[self.id_] = context.add_synth(
                    add_action=self.add_action,
                    permanent=False,
                    synthdef=self.synthdef or default,
                    target_node=self._resolve_target_node(
                        proxy_mapping,
                    ),
                    **settings,
                )
            else:
                if not isinstance(node := proxy_mapping.get(self.id_), Node):
                    raise RuntimeError(node)
                set_kwargs, map_kwargs = _partition_synth_settings(**settings)
                if set_kwargs:
                    context.set_node(node, **set_kwargs)
                if map_kwargs:
                    context.map_node(node, **map_kwargs)
            if self.duration:
                notes_mapping[self.id_] = current_offset + self.duration
        elif priority == Priority.STOP or self.rest:
            # check notes mapping for expected completion offset:
            # if expected completion >= current_offset or non-existent? bail
            # otherwise release
            expected_completion = notes_mapping.get(self.id_)
            if expected_completion is None:
                proxy_mapping.pop(self.id_, None)
            elif expected_completion > current_offset:
                return
            else:
                notes_mapping.pop(self.id_)
                if not isinstance(node := proxy_mapping.pop(self.id_), Node):
                    raise RuntimeError(node)
                context.free_node(node)


class NullEvent(Event):
    def expand(self, offset: float) -> Sequence[tuple[float, Priority, Event]]:
        return []


class SynthAllocateEvent(NodeEvent):
    def __init__(
        self,
        id_: UUID | tuple[UUID, int],
        synthdef: SynthDef,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 0.0,
        target_node: Node | UUID | None = None,
        **kwargs: SupportsFloat | UUID | str | Sequence[SupportsFloat | UUID | str],
    ) -> None:
        NodeEvent.__init__(
            self,
            add_action=add_action,
            delta=delta,
            id_=id_,
            target_node=target_node,
        )
        self.synthdef = synthdef
        self.kwargs = kwargs

    def perform(
        self,
        context: Context,
        proxy_mapping: dict[UUID | tuple[UUID, int], ContextObject],
        *,
        current_offset: float,
        notes_mapping: dict[UUID | tuple[UUID, int], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        settings = _process_synth_settings(proxy_mapping, **self.kwargs)
        # add the synth
        proxy_mapping[self.id_] = context.add_synth(
            add_action=self.add_action,
            permanent=False,
            synthdef=self.synthdef,
            target_node=self._resolve_target_node(
                proxy_mapping,
            ),
            **settings,
        )



import time
from typing import Generator, Literal

import pytest
from uqbar.strings import normalize

from supriya import Bus, BusGroup, Server, SynthDef, SynthDefBuilder
from supriya.ugens import Out, SinOsc


@pytest.fixture
def context(synthdef: SynthDef) -> Generator[Server, None, None]:
    context = Server().boot()
    with context.at():
        with context.add_synthdefs(synthdef):
            context.add_synth(synthdef=synthdef)
    context.sync()
    yield context
    context.quit()


@pytest.fixture
def synthdef() -> SynthDef:
    with SynthDefBuilder(frequency=440, out=0) as builder:
        Out.ar(bus=builder["out"], source=SinOsc.ar(frequency=builder["frequency"]))

    return builder.build(name="test:sine")


@pytest.mark.parametrize(
    "bus_index, expected_tree",
    [
        (
            0,
            """
            NODE TREE 0 group
                1 group
                    1000 test:sine
                        frequency: 440.0, out: 0.0
                1001 supriya:amp-scope-ar:1
                    in_: 0.0, max_frames: 4096.0, scope_frames: 4096.0, scope_id: 0.0
            """,
        ),
        (
            slice(0, 2),
            """
            NODE TREE 0 group
                1 group
                    1000 test:sine
                        frequency: 440.0, out: 0.0
                1001 supriya:amp-scope-ar:2
                    in_: 0.0, max_frames: 4096.0, scope_frames: 4096.0, scope_id: 0.0
            """,
        ),
    ],
)
def test_amplitude_scope(
    bus_index: int | slice, context: Server, expected_tree: str, synthdef: SynthDef
) -> None:
    bus_group = context.audio_output_bus_group
    buses: Bus | BusGroup
    if isinstance(bus_index, int):
        buses = bus_group[bus_index]
    else:
        start, stop, _ = bus_index.indices(len(bus_group))
        buses = BusGroup(
            calculation_rate=bus_group.calculation_rate,
            context=bus_group.context,
            count=stop - start,
            id_=start,
        )
    scope = context.add_amplitude_scope(bus=buses)
    context.sync()
    assert normalize(str(context.query_tree())) == normalize(expected_tree)
    results: list[tuple[int, list[float]]] = []
    for i in range(10):
        available_frames, data = scope.read()
        results.append((available_frames, data))
        print(
            f"{available_frames=} {data[:4]=} {scope.channel_count=} {scope.max_frames=}"
        )
        time.sleep(0.1)
    assert any([entry[0] for entry in results])
    assert all([any(entry[1]) for entry in results if entry[0]])
    scope.stop()


@pytest.mark.parametrize(
    "bus_index, fft_size, frequency_mode, expected_tree",
    [
        (
            0,
            4096,
            "linear",
            """
            NODE TREE 0 group
                1 group
                    1000 test:sine
                        frequency: 440.0, out: 0.0
                1001 supriya:freq-scope-lin-shm:1
                    fft_buffer_size: 4096.0, rate: 4.0, scope_id: 0.0, in_: 0.0
            """,
        ),
        # this one is "silent" because there's no output on the bus
        (
            4,
            8192,
            "linear",
            """
            NODE TREE 0 group
                1 group
                    1000 test:sine
                        frequency: 440.0, out: 0.0
                1001 supriya:freq-scope-lin-shm:1
                    fft_buffer_size: 8192.0, rate: 4.0, scope_id: 0.0, in_: 4.0
            """,
        ),
        (
            0,
            4096,
            "logarithmic",
            """
            NODE TREE 0 group
                1 group
                    1000 test:sine
                        frequency: 440.0, out: 0.0
                1001 supriya:freq-scope-log-shm:1
                    fft_buffer_size: 4096.0, rate: 4.0, scope_id: 0.0, in_: 0.0
            """,
        ),
    ],
)
def test_frequency_scope(
    bus_index: int,
    context: Server,
    expected_tree: str,
    fft_size: int,
    frequency_mode: Literal["linear", "logarithmic"],
    synthdef: SynthDef,
) -> None:
    bus_group = context.audio_output_bus_group
    scope = context.add_frequency_scope(
        bus=bus_group[bus_index],
        fft_size=fft_size,
        frequency_mode=frequency_mode,
    )
    context.sync()
    assert normalize(str(context.query_tree())) == normalize(expected_tree)
    results: list[tuple[int, list[float]]] = []
    for i in range(10):
        available_frames, data = scope.read()
        results.append((available_frames, data))
        print(
            f"{available_frames=} {data[:4]=} {scope.channel_count=} {scope.max_frames=}"
        )
        time.sleep(0.1)
    assert any([entry[0] for entry in results])
    assert all([any(entry[1]) for entry in results if entry[0]])
    scope.stop()



from uuid import UUID

import pytest

from supriya import AddAction, CalculationRate
from supriya.patterns import (
    BusAllocateEvent,
    BusFreeEvent,
    BusPattern,
    CompositeEvent,
    Event,
    EventPattern,
    GroupAllocateEvent,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    ParallelPattern,
    SequencePattern,
    SynthAllocateEvent,
)
from supriya.typing import CalculationRateLike
from supriya.ugens import system

from .conftest import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, inner_pattern, calculation_rate, channel_count, release_time, expected, is_infinite",
    [
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.0,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            UUID(int=0),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(UUID(int=1)),
                        SynthAllocateEvent(
                            UUID(int=2),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.0,
                            in_=UUID(int=0),
                            synthdef=system.system_link_audio_2,
                            target_node=UUID(int=1),
                        ),
                    ]
                ),
                NoteEvent(UUID(int=3), a=1, out=UUID(int=0), target_node=UUID(int=1)),
                NoteEvent(UUID(int=4), a=2, out=UUID(int=0), target_node=UUID(int=1)),
                CompositeEvent(
                    [
                        NodeFreeEvent(UUID(int=2)),
                        NodeFreeEvent(UUID(int=1)),
                        BusFreeEvent(UUID(int=0)),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            UUID(int=0),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(UUID(int=1)),
                        SynthAllocateEvent(
                            UUID(int=2),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=UUID(int=0),
                            synthdef=system.system_link_audio_2,
                            target_node=UUID(int=1),
                        ),
                    ]
                ),
                NoteEvent(UUID(int=3), a=1, out=UUID(int=0), target_node=UUID(int=1)),
                NoteEvent(UUID(int=4), a=2, out=UUID(int=0), target_node=UUID(int=1)),
                CompositeEvent(
                    [
                        NodeFreeEvent(UUID(int=2)),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(UUID(int=1)),
                        BusFreeEvent(UUID(int=0)),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            BusPattern(EventPattern(a=SequencePattern([1, 2])), channel_count=2),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            UUID(int=0),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(UUID(int=1)),
                        SynthAllocateEvent(
                            UUID(int=2),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=UUID(int=0),
                            synthdef=system.system_link_audio_2,
                            target_node=UUID(int=1),
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            UUID(int=3),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(UUID(int=4), target_node=UUID(int=1)),
                        SynthAllocateEvent(
                            UUID(int=5),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=UUID(int=3),
                            out=UUID(int=0),
                            synthdef=system.system_link_audio_2,
                            target_node=UUID(int=4),
                        ),
                    ]
                ),
                NoteEvent(UUID(int=6), a=1, out=UUID(int=3), target_node=UUID(int=4)),
                NoteEvent(UUID(int=7), a=2, out=UUID(int=3), target_node=UUID(int=4)),
                CompositeEvent(
                    [
                        NodeFreeEvent(UUID(int=5)),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(UUID(int=4)),
                        BusFreeEvent(UUID(int=3)),
                    ]
                ),
                CompositeEvent(
                    [
                        NodeFreeEvent(UUID(int=2)),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(UUID(int=1)),
                        BusFreeEvent(UUID(int=0)),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            ParallelPattern(
                [
                    BusPattern(
                        EventPattern(a=SequencePattern([1, 2])), channel_count=2
                    ),
                    BusPattern(
                        EventPattern(a=SequencePattern([1, 2])), channel_count=2
                    ),
                ]
            ),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            UUID(int=0),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(UUID(int=1)),
                        SynthAllocateEvent(
                            UUID(int=2),
                            system.system_link_audio_2,
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=UUID(int=0),
                            target_node=UUID(int=1),
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        CompositeEvent(
                            [
                                BusAllocateEvent(
                                    UUID(int=3),
                                    calculation_rate=CalculationRate.AUDIO,
                                    channel_count=2,
                                ),
                                GroupAllocateEvent(
                                    UUID(int=4), target_node=UUID(int=1)
                                ),
                                SynthAllocateEvent(
                                    UUID(int=5),
                                    system.system_link_audio_2,
                                    add_action=AddAction.ADD_AFTER,
                                    amplitude=1.0,
                                    fade_time=0.25,
                                    in_=UUID(int=3),
                                    out=UUID(int=0),
                                    target_node=UUID(int=4),
                                ),
                            ]
                        ),
                        NoteEvent(
                            UUID(int=6),
                            a=1,
                            delta=0.0,
                            out=UUID(int=3),
                            target_node=UUID(int=4),
                        ),
                        CompositeEvent(
                            [
                                BusAllocateEvent(
                                    UUID(int=7),
                                    calculation_rate=CalculationRate.AUDIO,
                                    channel_count=2,
                                ),
                                GroupAllocateEvent(
                                    UUID(int=8), target_node=UUID(int=1)
                                ),
                                SynthAllocateEvent(
                                    UUID(int=9),
                                    system.system_link_audio_2,
                                    add_action=AddAction.ADD_AFTER,
                                    amplitude=1.0,
                                    fade_time=0.25,
                                    in_=UUID(int=7),
                                    out=UUID(int=0),
                                    target_node=UUID(int=8),
                                ),
                            ]
                        ),
                        NoteEvent(
                            UUID(int=10),
                            a=1,
                            delta=0.0,
                            out=UUID(int=7),
                            target_node=UUID(int=8),
                        ),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        NoteEvent(
                            UUID(int=11),
                            a=2,
                            delta=0.0,
                            out=UUID(int=3),
                            target_node=UUID(int=4),
                        ),
                        NoteEvent(
                            UUID(int=12),
                            a=2,
                            delta=0.0,
                            out=UUID(int=7),
                            target_node=UUID(int=8),
                        ),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        CompositeEvent(
                            [
                                NodeFreeEvent(UUID(int=5)),
                                NullEvent(delta=0.25),
                                NodeFreeEvent(UUID(int=4)),
                                BusFreeEvent(UUID(int=3)),
                            ]
                        ),
                        CompositeEvent(
                            [
                                NodeFreeEvent(UUID(int=9)),
                                NullEvent(delta=0.25),
                                NodeFreeEvent(UUID(int=8)),
                                BusFreeEvent(UUID(int=7)),
                            ]
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        NodeFreeEvent(UUID(int=2)),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(UUID(int=1)),
                        BusFreeEvent(UUID(int=0)),
                    ]
                ),
            ],
            False,
        ),
    ],
)
def test_pattern(
    stop_at: float | None,
    inner_pattern,
    calculation_rate: CalculationRateLike,
    channel_count: int,
    release_time: float,
    expected: list[Event],
    is_infinite: bool,
) -> None:
    pattern = BusPattern(
        inner_pattern,
        calculation_rate=calculation_rate,
        channel_count=channel_count,
        release_time=release_time,
    )
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import pytest
from pytest_lazy_fixtures import lf
from uqbar.strings import normalize

from supriya import SynthDef, SynthDefBuilder, default
from supriya.ugens import FFT, MFCC, In, Out


@pytest.fixture
def mfcc_synthdef():
    with SynthDefBuilder() as builder:
        source = In.ar(bus=0)
        pv_chain = FFT.kr(source=source)
        mfcc = MFCC.kr(pv_chain=pv_chain)
        Out.kr(bus=0, source=mfcc)
    return builder.build("test:mfcc")


@pytest.mark.parametrize(
    "synthdef, expected_str",
    [
        (
            default,
            r"""
            digraph "synthdef_supriya:default" {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    ranksep=1,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    shape=Mrecord,
                    style="filled, rounded"];
                edge [penwidth=2];
                ugen_0 [fillcolor=lightsalmon2,
                    label="<f_0> Control\n(scalar) | { { <f_1_0_0> out:\n0.0 } }"];
                ugen_1 [fillcolor=lightgoldenrod2,
                    label="<f_0> Control\n(control) | { { <f_1_0_0> amplitude:\n0.1 | <f_1_0_1> frequency:\n440.0 | <f_1_0_2> gate:\n1.0 | <f_1_0_3> pan:\n0.5 } }"];
                ugen_2 [fillcolor=lightsteelblue2,
                    label="<f_0> VarSaw\n(audio) | { { <f_1_0_0> frequency | <f_1_0_1> initial_phase:\n0.0 | <f_1_0_2> width:\n0.3 } | { <f_1_1_0> 0 } }"];
                ugen_3 [fillcolor=lightgoldenrod2,
                    label="<f_0> Linen\n(control) | { { <f_1_0_0> gate | <f_1_0_1> attack_time:\n0.01 | <f_1_0_2> sustain_level:\n0.7 | <f_1_0_3> release_time:\n0.3 | <f_1_0_4> done_action:\n2.0 } | { <f_1_1_0> 0 } }"];
                ugen_4 [fillcolor=lightsalmon2,
                    label="<f_0> Rand\n(scalar) | { { <f_1_0_0> minimum:\n-0.4 | <f_1_0_1> maximum:\n0.0 } | { <f_1_1_0> 0 } }"];
                ugen_5 [fillcolor=lightgoldenrod2,
                    label="<f_0> BinaryOpUGen\n[ADDITION]\n(control) | { { <f_1_0_0> left | <f_1_0_1> right } | { <f_1_1_0> 0 } }"];
                ugen_6 [fillcolor=lightsteelblue2,
                    label="<f_0> VarSaw\n(audio) | { { <f_1_0_0> frequency | <f_1_0_1> initial_phase:\n0.0 | <f_1_0_2> width:\n0.3 } | { <f_1_1_0> 0 } }"];
                ugen_7 [fillcolor=lightsalmon2,
                    label="<f_0> Rand\n(scalar) | { { <f_1_0_0> minimum:\n0.0 | <f_1_0_1> maximum:\n0.4 } | { <f_1_1_0> 0 } }"];
                ugen_8 [fillcolor=lightgoldenrod2,
                    label="<f_0> BinaryOpUGen\n[ADDITION]\n(control) | { { <f_1_0_0> left | <f_1_0_1> right } | { <f_1_1_0> 0 } }"];
                ugen_9 [fillcolor=lightsteelblue2,
                    label="<f_0> VarSaw\n(audio) | { { <f_1_0_0> frequency | <f_1_0_1> initial_phase:\n0.0 | <f_1_0_2> width:\n0.3 } | { <f_1_1_0> 0 } }"];
                ugen_10 [fillcolor=lightsteelblue2,
                    label="<f_0> Sum3\n(audio) | { { <f_1_0_0> input_one | <f_1_0_1> input_two | <f_1_0_2> input_three } | { <f_1_1_0> 0 } }"];
                ugen_11 [fillcolor=lightsteelblue2,
                    label="<f_0> BinaryOpUGen\n[MULTIPLICATION]\n(audio) | { { <f_1_0_0> left | <f_1_0_1> right:\n0.3 } | { <f_1_1_0> 0 } }"];
                ugen_12 [fillcolor=lightsalmon2,
                    label="<f_0> Rand\n(scalar) | { { <f_1_0_0> minimum:\n4000.0 | <f_1_0_1> maximum:\n5000.0 } | { <f_1_1_0> 0 } }"];
                ugen_13 [fillcolor=lightsalmon2,
                    label="<f_0> Rand\n(scalar) | { { <f_1_0_0> minimum:\n2500.0 | <f_1_0_1> maximum:\n3200.0 } | { <f_1_1_0> 0 } }"];
                ugen_14 [fillcolor=lightgoldenrod2,
                    label="<f_0> XLine\n(control) | { { <f_1_0_0> start | <f_1_0_1> stop | <f_1_0_2> duration:\n1.0 | <f_1_0_3> done_action:\n0.0 } | { <f_1_1_0> 0 } }"];
                ugen_15 [fillcolor=lightsteelblue2,
                    label="<f_0> LPF\n(audio) | { { <f_1_0_0> source | <f_1_0_1> frequency } | { <f_1_1_0> 0 } }"];
                ugen_16 [fillcolor=lightsteelblue2,
                    label="<f_0> BinaryOpUGen\n[MULTIPLICATION]\n(audio) | { { <f_1_0_0> left | <f_1_0_1> right } | { <f_1_1_0> 0 } }"];
                ugen_17 [fillcolor=lightsteelblue2,
                    label="<f_0> BinaryOpUGen\n[MULTIPLICATION]\n(audio) | { { <f_1_0_0> left | <f_1_0_1> right } | { <f_1_1_0> 0 } }"];
                ugen_18 [fillcolor=lightsteelblue2,
                    label="<f_0> Pan2\n(audio) | { { <f_1_0_0> source | <f_1_0_1> position | <f_1_0_2> level:\n1.0 } | { <f_1_1_0> 0 | <f_1_1_1> 1 } }"];
                ugen_19 [fillcolor=lightsteelblue2,
                    label="<f_0> OffsetOut\n(audio) | { { <f_1_0_0> bus | <f_1_0_1> source[0] | <f_1_0_2> source[1] } }"];
                ugen_0:f_1_0_0:e -> ugen_19:f_1_0_0:w [color=salmon];
                ugen_1:f_1_0_0:e -> ugen_17:f_1_0_1:w [color=goldenrod];
                ugen_1:f_1_0_1:e -> ugen_2:f_1_0_0:w [color=goldenrod];
                ugen_1:f_1_0_1:e -> ugen_5:f_1_0_0:w [color=goldenrod];
                ugen_1:f_1_0_1:e -> ugen_8:f_1_0_0:w [color=goldenrod];
                ugen_1:f_1_0_2:e -> ugen_3:f_1_0_0:w [color=goldenrod];
                ugen_1:f_1_0_3:e -> ugen_18:f_1_0_1:w [color=goldenrod];
                ugen_2:f_1_1_0:e -> ugen_10:f_1_0_0:w [color=steelblue];
                ugen_3:f_1_1_0:e -> ugen_16:f_1_0_1:w [color=goldenrod];
                ugen_4:f_1_1_0:e -> ugen_5:f_1_0_1:w [color=salmon];
                ugen_5:f_1_1_0:e -> ugen_6:f_1_0_0:w [color=goldenrod];
                ugen_6:f_1_1_0:e -> ugen_10:f_1_0_1:w [color=steelblue];
                ugen_7:f_1_1_0:e -> ugen_8:f_1_0_1:w [color=salmon];
                ugen_8:f_1_1_0:e -> ugen_9:f_1_0_0:w [color=goldenrod];
                ugen_9:f_1_1_0:e -> ugen_10:f_1_0_2:w [color=steelblue];
                ugen_10:f_1_1_0:e -> ugen_11:f_1_0_0:w [color=steelblue];
                ugen_11:f_1_1_0:e -> ugen_15:f_1_0_0:w [color=steelblue];
                ugen_12:f_1_1_0:e -> ugen_14:f_1_0_0:w [color=salmon];
                ugen_13:f_1_1_0:e -> ugen_14:f_1_0_1:w [color=salmon];
                ugen_14:f_1_1_0:e -> ugen_15:f_1_0_1:w [color=goldenrod];
                ugen_15:f_1_1_0:e -> ugen_16:f_1_0_0:w [color=steelblue];
                ugen_16:f_1_1_0:e -> ugen_17:f_1_0_0:w [color=steelblue];
                ugen_17:f_1_1_0:e -> ugen_18:f_1_0_0:w [color=steelblue];
                ugen_18:f_1_1_0:e -> ugen_19:f_1_0_1:w [color=steelblue];
                ugen_18:f_1_1_1:e -> ugen_19:f_1_0_2:w [color=steelblue];
            }
            """,
        ),
        (
            lf("mfcc_synthdef"),
            r"""
            digraph "synthdef_test:mfcc" {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    ranksep=1,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    shape=Mrecord,
                    style="filled, rounded"];
                edge [penwidth=2];
                ugen_0 [fillcolor=lightsteelblue2,
                    label="<f_0> In\n(audio) | { { <f_1_0_0> bus:\n0.0 } | { <f_1_1_0> 0 } }"];
                ugen_1 [fillcolor=lightsalmon2,
                    label="<f_0> MaxLocalBufs\n(scalar) | { { <f_1_0_0> maximum:\n1.0 } | { <f_1_1_0> 0 } }"];
                ugen_2 [fillcolor=lightsalmon2,
                    label="<f_0> LocalBuf\n(scalar) | { { <f_1_0_0> channel_count:\n1.0 | <f_1_0_1> frame_count:\n2048.0 } | { <f_1_1_0> 0 } }"];
                ugen_3 [fillcolor=lightgoldenrod2,
                    label="<f_0> FFT\n(control) | { { <f_1_0_0> buffer_id | <f_1_0_1> source | <f_1_0_2> hop:\n0.5 | <f_1_0_3> window_type:\n0.0 | <f_1_0_4> active:\n1.0 | <f_1_0_5> window_size:\n0.0 } | { <f_1_1_0> 0 } }"];
                ugen_4 [fillcolor=lightgoldenrod2,
                    label="<f_0> MFCC\n(control) | { { <f_1_0_0> pv_chain | <f_1_0_1> coeff_count:\n13.0 } | { <f_1_1_0> 0 | <f_1_1_1> 1 | <f_1_1_2> 2 | <f_1_1_3> 3 | <f_1_1_4> 4 | <f_1_1_5> 5 | <f_1_1_6> 6 | <f_1_1_7> 7 | <f_1_1_8> 8 | <f_1_1_9> 9 | <f_1_1_10> 10 | <f_1_1_11> 11 | <f_1_1_12> 12 } }"];
                ugen_5 [fillcolor=lightgoldenrod2,
                    label="<f_0> Out\n(control) | { { <f_1_0_0> bus:\n0.0 | <f_1_0_1> source[0] | <f_1_0_2> source[1] | <f_1_0_3> source[2] | <f_1_0_4> source[3] | <f_1_0_5> source[4] | <f_1_0_6> source[5] | <f_1_0_7> source[6] | <f_1_0_8> source[7] | <f_1_0_9> source[8] | <f_1_0_10> source[9] | <f_1_0_11> source[10] | <f_1_0_12> source[11] | <f_1_0_13> source[12] } }"];
                ugen_0:f_1_1_0:e -> ugen_3:f_1_0_1:w [color=steelblue];
                ugen_2:f_1_1_0:e -> ugen_3:f_1_0_0:w [color=salmon];
                ugen_3:f_1_1_0:e -> ugen_4:f_1_0_0:w [color=goldenrod];
                ugen_4:f_1_1_0:e -> ugen_5:f_1_0_1:w [color=goldenrod];
                ugen_4:f_1_1_1:e -> ugen_5:f_1_0_2:w [color=goldenrod];
                ugen_4:f_1_1_2:e -> ugen_5:f_1_0_3:w [color=goldenrod];
                ugen_4:f_1_1_3:e -> ugen_5:f_1_0_4:w [color=goldenrod];
                ugen_4:f_1_1_4:e -> ugen_5:f_1_0_5:w [color=goldenrod];
                ugen_4:f_1_1_5:e -> ugen_5:f_1_0_6:w [color=goldenrod];
                ugen_4:f_1_1_6:e -> ugen_5:f_1_0_7:w [color=goldenrod];
                ugen_4:f_1_1_7:e -> ugen_5:f_1_0_8:w [color=goldenrod];
                ugen_4:f_1_1_8:e -> ugen_5:f_1_0_9:w [color=goldenrod];
                ugen_4:f_1_1_9:e -> ugen_5:f_1_0_10:w [color=goldenrod];
                ugen_4:f_1_1_10:e -> ugen_5:f_1_0_11:w [color=goldenrod];
                ugen_4:f_1_1_11:e -> ugen_5:f_1_0_12:w [color=goldenrod];
                ugen_4:f_1_1_12:e -> ugen_5:f_1_0_13:w [color=goldenrod];
            }
            """,
        ),
    ],
)
def test_SynthDef___graph__(synthdef: SynthDef, expected_str: str) -> None:
    graph = synthdef.__graph__()
    actual_str = format(graph, "graphviz")
    assert normalize(actual_str) == normalize(expected_str)



"""
Exceptions.
"""


class AllocationError(Exception):
    pass


class ContextError(Exception):
    pass


class InvalidCalculationRate(ContextError):
    pass


class InvalidMoment(ContextError):
    pass


class MomentClosed(ContextError):
    pass


class ServerCannotBoot(Exception):
    pass


class ServerOnline(Exception):
    pass


class ServerOffline(Exception):
    pass


class OwnedServerShutdown(Exception):
    pass


class UnownedServerShutdown(Exception):
    pass


class TooManyClients(Exception):
    pass



import pytest
from pytest_lazy_fixtures import lf
from uqbar.strings import normalize

from supriya.ugens import Envelope, EnvGen, Out, SynthDef, SynthDefBuilder


@pytest.fixture
def envelope_a() -> SynthDef:
    with SynthDefBuilder(out=0, decay=1) as builder:
        envelope = Envelope([0.11, 1, 0], [0, builder["decay"]], [-225])
        envgen = EnvGen.kr(envelope=envelope)
        Out.kr(bus=builder["out"], source=envgen)
    return builder.build(name="test-envelope-expansion")


@pytest.fixture
def envelope_b() -> SynthDef:
    with SynthDefBuilder(out=0, decay=[1, 2]) as builder:
        envelope = Envelope([0.11, 1, 0], [0, builder["decay"]], [-225])
        envgen = EnvGen.kr(envelope=envelope)
        Out.kr(bus=builder["out"], source=envgen)
    return builder.build(name="test-envelope-expansion")


@pytest.mark.parametrize(
    "synthdef, expected_str",
    [
        (
            lf("envelope_a"),
            """
            synthdef:
                name: test-envelope-expansion
                ugens:
                -   Control.kr:
                        decay: 1.0
                        out: 0.0
                -   EnvGen.kr:
                        gate: 1.0
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: 1.0
                        done_action: 0.0
                        envelope[0]: 0.11
                        envelope[1]: 2.0
                        envelope[2]: -99.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 0.0
                        envelope[6]: 5.0
                        envelope[7]: -225.0
                        envelope[8]: 0.0
                        envelope[9]: Control.kr[0:decay]
                        envelope[10]: 5.0
                        envelope[11]: -225.0
                -   Out.kr:
                        bus: Control.kr[1:out]
                        source[0]: EnvGen.kr[0]
            """,
        ),
        (
            lf("envelope_b"),
            """
            synthdef:
                name: test-envelope-expansion
                ugens:
                -   Control.kr:
                        decay[0]: 1.0
                        decay[1]: 2.0
                        out: 0.0
                -   EnvGen.kr/0:
                        gate: 1.0
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: 1.0
                        done_action: 0.0
                        envelope[0]: 0.11
                        envelope[1]: 2.0
                        envelope[2]: -99.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 0.0
                        envelope[6]: 5.0
                        envelope[7]: -225.0
                        envelope[8]: 0.0
                        envelope[9]: Control.kr[0:decay[0]]
                        envelope[10]: 5.0
                        envelope[11]: -225.0
                -   EnvGen.kr/1:
                        gate: 1.0
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: 1.0
                        done_action: 0.0
                        envelope[0]: 0.11
                        envelope[1]: 2.0
                        envelope[2]: -99.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 0.0
                        envelope[6]: 5.0
                        envelope[7]: -225.0
                        envelope[8]: 0.0
                        envelope[9]: Control.kr[1:decay[1]]
                        envelope[10]: 5.0
                        envelope[11]: -225.0
                -   Out.kr:
                        bus: Control.kr[2:out]
                        source[0]: EnvGen.kr/0[0]
                        source[1]: EnvGen.kr/1[0]
            """,
        ),
    ],
)
def test_Envelope_serialize(synthdef: SynthDef, expected_str: str):
    """
    Envelopes shouldn't expand trivial UGenOperable inputs.
    """
    assert str(synthdef) == normalize(expected_str)



import os
import stat
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

from supriya import scsynth


@pytest.fixture
def mock_env_scsynth_path(monkeypatch):
    monkeypatch.delenv(scsynth.ENVAR_SERVER_EXECUTABLE, raising=False)
    monkeypatch.setenv("PATH", "")


def test_find_via_argument(mock_env_scsynth_path):
    with NamedTemporaryFile() as tmp:
        expected = Path(tmp.name).absolute()
        expected.chmod(expected.stat().st_mode | stat.S_IEXEC)
        actual = scsynth.find(expected)
        assert actual == expected


def test_find_via_env_var(mock_env_scsynth_path, monkeypatch):
    with NamedTemporaryFile() as tmp:
        expected = Path(tmp.name).absolute()
        expected.chmod(expected.stat().st_mode | stat.S_IEXEC)
        monkeypatch.setenv(scsynth.ENVAR_SERVER_EXECUTABLE, str(expected))
        actual = scsynth.find()
        assert actual == expected


def test_find_on_path(mock_env_scsynth_path, monkeypatch):
    with TemporaryDirectory() as tmp_dir:
        scsynth_path = Path(tmp_dir) / "scsynth"
        with open(scsynth_path, "w"):
            scsynth_path.chmod(scsynth_path.stat().st_mode | stat.S_IEXEC)
            monkeypatch.setenv("PATH", os.pathsep + tmp_dir)
            actual = scsynth.find()
            expected = Path("scsynth")
            assert actual == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        # nothing specified
        (
            {},
            [
                "/path/to/executable",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # only input device defined
        (
            {"input_device": "Device X"},
            [
                "/path/to/executable",
                "-H",
                "Device X",
                "",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # only output device defined
        (
            {"output_device": "Device Y"},
            [
                "/path/to/executable",
                "-H",
                "",
                "Device Y",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # input and output devices defined, but different
        (
            {"input_device": "Device P", "output_device": "Device Q"},
            [
                "/path/to/executable",
                "-H",
                "Device P",
                "Device Q",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # input and output devices defined, and identical
        (
            {"input_device": "Device Z", "output_device": "Device Z"},
            [
                "/path/to/executable",
                "-H",
                "Device Z",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # many flags specified
        (
            {
                "audio_bus_channel_count": 1024 * 2,
                "block_size": 64 * 2,
                "buffer_count": 1024 * 2,
                "control_bus_channel_count": 16384 * 2,
                "executable": "supernova",
                "hardware_buffer_size": 512,
                "initial_node_id": 100,
                "input_bus_channel_count": 16,
                "input_device": "Hypothetical Input Device",
                "input_stream_mask": "01101010",
                "ip_address": "0.0.0.0",
                "load_synthdefs": False,
                "maximum_logins": 23,
                "maximum_node_count": 1024 * 2,
                "maximum_synthdef_count": 1024 * 2,
                "memory_locking": True,
                "memory_size": 8192 * 2,
                "output_bus_channel_count": 8 * 2,
                "output_device": "Hypothetical Output Device",
                "output_stream_mask": "10010101",
                "password": "changeme",
                "protocol": "tcp",
                "random_number_generator_count": 64 * 2,
                "restricted_path": "/restricted/path",
                "safety_clip": "inf",
                "sample_rate": 44100,
                "threads": 8,
                "ugen_plugins_path": "/path/to/ugens",
                "verbosity": 1,
                "wire_buffer_count": 64 * 22,
                "zero_configuration": True,
            },
            [
                "/path/to/executable",
                "-B",
                "0.0.0.0",
                "-D",
                "0",
                "-H",
                "Hypothetical Input Device",
                "Hypothetical Output Device",
                "-I",
                "01101010",
                "-L",
                "-O",
                "10010101",
                "-P",
                "/restricted/path",
                "-S",
                "44100",
                "-T",
                "8",
                "-U",
                "/path/to/ugens",
                "-Z",
                "512",
                "-a",
                "2048",
                "-b",
                "2048",
                "-c",
                "32768",
                "-d",
                "2048",
                "-i",
                "16",
                "-l",
                "23",
                "-m",
                "16384",
                "-n",
                "2048",
                "-o",
                "16",
                "-p",
                "changeme",
                "-r",
                "128",
                "-s",
                "inf",
                "-t",
                "57110",
                "-v",
                "1",
                "-w",
                "1408",
                "-z",
                "128",
            ],
        ),
    ],
)
def test_Options(kwargs: dict, expected: list[str]) -> None:
    options = scsynth.Options(**kwargs)
    actual = list(options)
    actual[0] = "/path/to/executable"  # replace to make it portable
    assert actual == expected
